from typing import List, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .transition_rule import TransitionRule

class Encounter(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  previousEncounterId: Union[str, None] = None
  nextEncounterId: Union[str, None] = None
  encounterScheduledAtTimingId: Union[str, None] = None
  encounterEnvironmentalSetting: Union[Code, None] = None
  encounterContactModes: List[Code] = []
  transitionStartRule: Union[TransitionRule, None] = None
  transitionEndRule: Union[TransitionRule, None] = None



import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.study_design_activity_sheet.study_design_activity_sheet import StudyDesignActivitySheet
from usdm_model.code import Code

def test_create(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['ActivityId_1', 'ActivityId_2', 'ActivityId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Activity 1', 'Activity One', '', ''], ['Activity 2', 'Activity Two', 'T', 'Condition 2'], ['Activity 3', 'Activity Three', 'F', '']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['activityName', 'activityDescription', 'activityIsConditional', 'activityIsConditionalReason'])
  activities = StudyDesignActivitySheet("")
  assert len(activities.items) == 3
  assert activities.items[0].id == 'ActivityId_1'
  assert activities.items[0].name == 'Activity 1'
  assert activities.items[0].description == 'Activity One'
  assert activities.items[0].activityIsConditional == False
  assert activities.items[0].activityIsConditionalReason == ''
  assert activities.items[1].id == 'ActivityId_2'
  assert activities.items[1].description == 'Activity Two'
  assert activities.items[1].activityIsConditional == True
  assert activities.items[1].activityIsConditionalReason == 'Condition 2'
  assert activities.items[2].id == 'ActivityId_3'
  assert activities.items[2].activityIsConditional == False
  
def test_create_empty(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['studyActivityName', 'studyActivityDescription', 'studyActivityType'])
  activities = StudyDesignActivitySheet("")
  assert len(activities.items) == 0

def test_read_cell_by_name_error(mocker):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['Activity 1', 'Activity One']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['activityName', 'activityDescription'])
  activities = StudyDesignActivitySheet("")
  mock_error.assert_called()
  assert call_parameters == [
    ("studyDesignActivities", 1, -1, "Error reading cell 'activityIsConditional'", 10),
    ("studyDesignActivities", 1, -1, "Error reading cell 'activityIsConditionalReason'", 10),
  ]
  



from typing import Union
from .api_base_model import ApiBaseModelWithId
from .code import Code

class Address(ApiBaseModelWithId):
  text: str = ""
  line: str = ""
  city: str = ""
  district: str = ""
  state: str = ""
  postalCode: str = ""
  country: Union[Code, None] = None



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary

import traceback

class StudyDesignDictionarySheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='dictionaries', optional=True)
      self.items = []
      if self.success:
        current_name = None
        current_dictionary = None
        current_map = {}
        for index, row in self.sheet.iterrows():
          name = self.read_cell_by_name(index, 'name')
          if name:
            if name != current_name:
              current_name = name
              if current_dictionary:
                current_dictionary.parameterMap = current_map
                current_map = {}
              description = self.read_cell_by_name(index, 'description')
              label = self.read_cell_by_name(index, 'label')
              current_dictionary = self._dictionary(name, description, label)
          key = self.read_cell_by_name(index, 'key')
          klass = self.read_cell_by_name(index, 'class')
          xref_name = self.read_cell_by_name(index, 'xref')
          attribute = self.read_cell_by_name(index, 'attribute')
          item = cross_references.get(klass, xref_name)
          if item:
            current_map[key] = {'klass': klass, 'id': item.id, 'attribute': attribute}
          else:
            self._general_warning(f"Unable to resolve dictionary reference klass: '{klass}', name: '{xref_name}', attribute '{attribute}'")
        # Clean up last dictionary if present
        if current_dictionary:
          current_dictionary.parameterMap = current_map
        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _dictionary(self, name, description, label):
    try:
      item = SyntaxTemplateDictionary(
        id=id_manager.build_id(SyntaxTemplateDictionary), 
        name=name,
        description=description,
        label=label,
        parameterMap={}
      )
    except Exception as e:
      self._general_error(f"Failed to create SyntaxTemplateDictionary object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      return None
    else:
      self.items.append(item)
      cross_references.add(name, item)
      return item



from typing import List, Dict, Union, Literal
from .api_base_model import ApiBaseModelWithId
from .timing import Timing

class ScheduledInstance(ApiBaseModelWithId):
  instanceType: Literal['ACTIVITY', 'DECISION']
  scheduledInstanceTimings: List[Timing] = []
  scheduledInstanceTimelineId: Union[str, None] = None
  scheduleTimelineExitId: Union[str, None] = None
  defaultConditionId: Union[str, None] = None
  epochId: Union[str, None] = None

class ScheduledActivityInstance(ScheduledInstance):
  activityIds: List[str] = []
  scheduledActivityInstanceEncounterId: Union[str, None] = None

class ScheduledDecisionInstance(ScheduledInstance):
  conditionAssignments: List[List] = []



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.study_amendment import StudyAmendment
from usdm_model.study_amendment_reason import StudyAmendmentReason
from usdm_model.subject_enrollment import SubjectEnrollment
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.iso_3166 import ISO3166
from usdm_excel.alias import Alias

import traceback

class StudyDesignAmendmentSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyAmendments', optional=True)
      self.items = []
      if self.success:
        for index, row in self.sheet.iterrows():
          secondaries = []
          number = self.read_cell_by_name(index, 'number')
          summary = self.read_description_by_name(index, 'summary')
          substantial = self.read_boolean_cell_by_name(index, 'substantialImpact')
          primary_reason = self._read_primary_reason_cell(index)
          primary = self._amendment_reason(primary_reason)
          #print(f"PRIMARY: {primary}")
          secondary_reasons = self._read_secondary_reason_cell(index)
          for reason in secondary_reasons:
            amendment_reason = self._amendment_reason(reason)        
            if amendment_reason:
              #print(f"SECONDARY: {amendment_reason}")
              secondaries.append(amendment_reason)
          enrollment = self._read_enrollment_cell(index)
          enrollments = self._enrollments(enrollment)
          #print(f"ENROLLMENT: {enrollment}")
          try:
            item = StudyAmendment(
              id=id_manager.build_id(StudyAmendment), 
              number=number,
              summary=summary,
              substantialImpact=substantial,
              primaryReason=primary,
              secondaryReasons=secondaries,
              enrollments=enrollments
            )
          except Exception as e:
            self._general_error(f"Failed to create StudyAmendment object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            self.items.append(item)
            cross_references.add(item.id, item)
        self.items.sort(key=lambda d: int(d.number))
        self.previous_link(self.items, 'previousId')
        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _enrollments(self, enrollments):
    results = []
    for enrollment in enrollments:
      try:
        item = SubjectEnrollment(
          id=id_manager.build_id(SubjectEnrollment), 
          type=enrollment['type'],
          code=enrollment['code'],
          quantity=enrollment['quantity']
        )
      except Exception as e:
        self._general_error(f"Failed to create SubjectEnrollment object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")
        return None
      else:
        results.append(item)
    return results

  def _amendment_reason(self, reason):
    #print(f"AR1: {reason}")
    try:
      item = StudyAmendmentReason(
        id=id_manager.build_id(StudyAmendmentReason), 
        code=reason['code'],
        otherReason=reason['other']
      )
    except Exception as e:
      self._general_error(f"Failed to create StudyAmendmentReason object, exception {e}")
      self._traceback(f"{traceback.format_exc()}")
      print(f"AR2: {traceback.format_exc()}")
      return None
    else:
      return item
    
  def _read_enrollment_cell(self, row_index):
    result = []
    col_index = self.sheet.columns.get_loc('enrollment')
    value = self.read_cell(row_index, col_index)
    #print(f"ENROL1: {value}")
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where geographic enrollment values expected")
      return [{'type': CDISCCT().code_for_attribute('GeographicScope', 'type', 'Global'), 'code': None, 'quantity': '0'}]
    else:
      for item in self._state_split(value):
        #print(f"ENROL2: {item}")
        if item.strip().upper().startswith("GLOBAL"):
          # If we ever find global just return the one code
          text = item.strip()
          parts = text.split(":")
          if len(parts) == 2:
            return [{'type': CDISCCT().code_for_attribute('GeographicScope', 'type', 'Global'), 'code': None, 'quantity': parts[1].strip()}]
          else:
            self._error(row_index, col_index, f"Failed to decode enrollment data {item}, no '=' detected")
          return 
        else: 
          code = None
          if item.strip():
            outer_parts = item.split(":")
            if len(outer_parts) == 2:
              system = outer_parts[0].strip()
              value = outer_parts[1].strip()
              name_value = value.split('=')
              if len(name_value) == 2:
                quantity = name_value[1].strip()
                if system.upper() == "REGION":
                  pt = 'Region'
                  code = ISO3166().region_code(value)
                elif system.upper() == "COUNTRY":
                  pt = 'Country'
                  code = ISO3166().code(value)
                else:
                  self._error(row_index, col_index, f"Failed to decode geographic enrollment data {name_value}, must be either Region or Country using ISO3166 codes")
              else:
                self._error(row_index, col_index, f"Failed to decode geographic enrollment data {name_value}, no '=' detected")
            else:
              self._error(row_index, col_index, f"Failed to decode geographic enrollment data {outer_parts}, no ':' detected")
          else:
            self._error(row_index, col_index, f"Failed to decode geographic enrollment data {item}, appears empty")
          if code:
            result.append({'type': CDISCCT().code_for_attribute('GeographicScope', 'type', pt), 'code': Alias().code(code, []), 'quantity': quantity})
      return result


  def _read_secondary_reason_cell(self, row_index):
    results = []
    col_index = self.sheet.columns.get_loc('secondaryReasons')
    value = self.read_cell(row_index, col_index)
    if not value.strip():
      return results
    parts = value.strip().split(',')
    for part in parts:
      result = self._extract_reason(part, row_index, col_index)
      if result:
        results.append(result)
    return results
      
  def _read_primary_reason_cell(self, row_index):
    col_index = self.sheet.columns.get_loc('primaryReason')
    value = self.read_cell(row_index, col_index)
    return self._extract_reason(value, row_index, col_index)
  
  def _extract_reason(self, value, row_index, col_index):
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where CDISC CT value expected.")
      return None
    elif value.strip().upper().startswith('OTHER'):
      text = value.strip()
      parts = text.split("=")
      if len(parts) == 2:
        return {'code': CDISCCT().code_for_attribute('StudyAmendmentReason', 'code', 'Other'), 'other': parts[1].strip()}
      else:
        self._error(row_index, col_index, f"Failed to decode reason data {text}, no '=' detected")
    else:
      code = CDISCCT().code_for_attribute('StudyAmendmentReason', 'code', value)
      if code is None:
        self._error(row_index, col_index, f"CDISC CT not found for value '{value}'.")
      return {'code': code, 'other': None}


import shlex
import pandas as pd
from openpyxl import load_workbook
from usdm_excel.id_manager import id_manager
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.other_ct import OtherCT
from usdm_excel.ct_version_manager import ct_version_manager
from usdm_excel.errors.errors import error_manager
from usdm_excel.logger import package_logger
from usdm_excel.option_manager import *

class BaseSheet():

  class StateError(Exception):
    pass

  class FormatError(Exception):
    pass

  def __init__(self, file_path, sheet_name, header=0, optional=False, converters={}, require={}):
    self.file_path = file_path
    self.sheet_name = sheet_name
    self.sheet = None
    self.success = False
    if optional and not self._sheet_present(file_path, sheet_name):
      self._general_info(f"{sheet_name} not found but optional")
    else:
      if require and not self._check_cell_value(file_path, sheet_name, require['row'], require['column'], require['value']):
        self._general_info(f"Required value {require['value']} at [{require['row']}, {require['column']}] mismatch in {sheet_name}")
        pass
      else:
        self.sheet = pd.read_excel(open(file_path, 'rb'), sheet_name=sheet_name, header=header, converters=converters)
        self.success = True
        self._general_info("Processed sheet %s" % (sheet_name))

  def cell_empty(self, row_index, col_index):
    return pd.isnull(self.sheet.iloc[row_index, col_index])  

  def read_cell_by_name(self, row_index, field_name, default=None):
    try:
      col_index = self.column_present(field_name)
      return self.read_cell(row_index, col_index)
    except Exception as e:
      if default is not None:
        return default
      else:
        col_index = -2
        self._error(row_index, col_index, f"Error reading cell '{field_name}'")
        return ""

  def column_present(self, names):
    fields = [names] if isinstance(names, str) else names
    for field in fields:
      try:
        col_index = self.sheet.columns.get_loc(field)
        return col_index
      except:
        pass
    raise BaseSheet.FormatError(f"Failed to detect column(s) '{', '.join(fields)}' in sheet")

  def read_cell(self, row_index, col_index):
    try:
      if pd.isnull(self.sheet.iloc[row_index, col_index]):
        return ""
      else:
        return str(self.sheet.iloc[row_index, col_index]).strip()
    except Exception as e:
      self._error(row_index, col_index, "Error (%s) reading cell" % (e))
      return ""

  def read_cell_empty_legacy(self, row_index, col_index):
    if self.cell_empty(row_index, col_index):
      return "", True
    else:
      value = self.read_cell_empty(row_index, col_index, '-')
      if value == "":
        return "", True
      else:
        return value, False

  def read_cell_empty(self, row_index, col_index, empty_character):
    value = self.read_cell(row_index, col_index)
    value = "" if value == empty_character else value
    return value

  def read_cell_multiple(self, rindex, cindex):
    try:
      results = []
      value = self.read_cell(rindex, cindex)
      if value.strip() == '':
        return results
      for part in self._state_split(value):
        results.append(part.strip())
      return results
    except BaseSheet.StateError as e:
      self._error(rindex, cindex, f"Internal state error ({e}) reading cell")
    except BaseSheet.FormatError as e:
      self._error(rindex, cindex, "Format error reading cell, check the format of the cell")

  def read_cell_with_previous(self, row_index, col_index, first_col_index):
    try:
      i = col_index
      while i >= first_col_index:
        if pd.isnull(self.sheet.iloc[row_index, i]):
          i -= 1
        else:
          return self.sheet.iloc[row_index, i].strip()
      self._warning(row_index, col_index, "No previous non-empty cell found.")
      return ""
    except Exception as e:
      self._error(row_index, col_index, "Error (%s) reading cell row '%s', field '%s'" % (e, row_index, col_index))
      return ""

  def read_boolean_cell_by_name(self, row_index, field_name):
    value = self.read_cell_by_name(row_index, field_name)
    if value.strip().upper() in ['Y', 'YES', 'T', 'TRUE', '1']:
      return True
    return False

  def read_description_by_name(self, row_index, field_name):
    value = self.read_cell_by_name(row_index, field_name)
    empty_value = option_manager.get(Options.DESCRIPTION)
    if value == "" and not empty_value == "":
      return empty_value
    return value

  # Want to kill this method
  def set_cdisc_code(self, value):
    if value.strip() == "":
      return None
    parts = value.split("=")
    try:
      return CDISCCT().code(code=parts[0].strip(), decode=parts[1].strip())
    except Exception as e:
      self._error(0, 0, "CDISC code error '%s'" % (e))
      return None

  def read_other_code_cell_by_name(self, row_index, field_name):
    #col_index = self.sheet.columns.get_loc(field_name)
    col_index = self.column_present(field_name)
    return self.read_other_code_cell(row_index, col_index)

  def read_other_code_cell(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      return None
    return self._decode_other_code(value, row_index, col_index)

  def read_other_code_cell_multiple_by_name(self, row_index, field_name):
    col_index = self.sheet.columns.get_loc(field_name)
    return self.read_other_code_cell_mutiple(row_index, col_index)

  def read_other_code_cell_mutiple(self, row_index, col_index):
    value = self.read_cell(row_index, col_index)
    result = []
    if value.strip() == '':
      return result
    for item in self._state_split(value):
      code = self._decode_other_code(item.strip(), row_index, col_index)
      if not code == None:
        result.append(code)
    return result

  def read_cdisc_klass_attribute_cell_by_name(self, klass, attribute, row_index, field_name):
    #col_index = self.sheet.columns.get_loc(field_name)
    col_index = self.column_present(field_name)
    return self.read_cdisc_klass_attribute_cell(klass, attribute, row_index, col_index)

  def read_cdisc_klass_attribute_cell(self, klass, attribute, row_index, col_index):
    code = None
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where CDISC CT value expected.")
    else:
      code = CDISCCT().code_for_attribute(klass, attribute, value)
      if code is None:
        self._error(row_index, col_index, f"CDISC CT not found for value '{value}'.")
    return code

  def read_cdisc_klass_attribute_cell_multiple_by_name(self, klass, attribute, row_index, field_name):
    #col_index = self.sheet.columns.get_loc(field_name)
    col_index = self.column_present(field_name)
    return self.read_cdisc_klass_attribute_cell_multiple(klass, attribute, row_index, col_index)

  def read_cdisc_klass_attribute_cell_multiple(self, klass, attribute, row_index, col_index):
    result = []
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where multiple CDISC CT values expected.")
      return result
    for item in self._state_split(value):
      code = CDISCCT().code_for_attribute(klass, attribute, item.strip())
      if code is not None:
        result.append(code)
      else:
        self._error(row_index, col_index, f"CDISC CT not found for value '{item.strip()}'.")
    return result

  def double_link(self, items, prev, next):
    try: 
      for idx, item in enumerate(items):
        if idx == 0:
          if option_manager.get(Options.PREVIOUS_NEXT) == PrevNextOption.NULL_STRING.value:
            setattr(item, prev, "")
          else:
            setattr(item, prev, None)
        else:
          the_id = getattr(items[idx-1], 'id')
          setattr(item, prev, the_id)
        if idx == len(items)-1:  
          if option_manager.get(Options.PREVIOUS_NEXT) == PrevNextOption.NULL_STRING.value:
            setattr(item, next, "")
          else:
            setattr(item, next, None)
        else:
          the_id = getattr(items[idx+1], 'id')
          setattr(item, next, the_id)
    except Exception as e:
      self._general_error(f"Exception {e} in double_link: {items}")

  def previous_link(self, items, prev):
    try: 
      for idx, item in enumerate(items):
        if idx == 0:
          if option_manager.get(Options.PREVIOUS_NEXT) == PrevNextOption.NULL_STRING.value:
            setattr(item, prev, "")
          else:
            setattr(item, prev, None)
        else:
          the_id = getattr(items[idx-1], 'id')
          setattr(item, prev, the_id)
    except Exception as e:
      self._general_error(f"Exception {e} in previous_link: {items}")

  def _decode_other_code(self, value, row_index, col_index):
    if value.strip() == "":
      return None
    outer_parts = value.split(":")
    if len(outer_parts) == 2:
      system = outer_parts[0].strip()
      inner_parts = outer_parts[1].strip().split("=")
      if len(inner_parts) == 2:
        version = ct_version_manager.get(system)
        return OtherCT().code(code=inner_parts[0].strip(), system=system, version=version, decode=inner_parts[1].strip())
      else:
        self._error(row_index, col_index, "Failed to decode code data '%s', no '=' detected" % (value))
    else:
      self._error(row_index, col_index, "Failed to decode code data '%s', no ':' detected" % (value))
    return None

  def _to_int(self, value):
    try:
      return int(value)
    except:
      return None

  def _add_errors(self, errors, row, column):
    for error in errors:
      self._error(row, column, error)
      
  def _get_column_index(self, column_name):
    return self.sheet.columns.get_loc(column_name)

  def _info(self, row, column, message):
     package_logger.info(self._format(row + 1, column + 1, message))
     
  def _general_info(self, message):
     package_logger.info(self._format(None, None, message))
     
  def _error(self, row, column, message):
    error_manager.add(self.sheet_name, row + 1, column + 1, message)

  def _general_error(self, message):
    error_manager.add(self.sheet_name, None, None, message)

  def _warning(self, row, column, message):
    error_manager.add(self.sheet_name, row + 1, column + 1, message, error_manager.WARNING)

  def _general_warning(self, message):
    error_manager.add(self.sheet_name, None, None, message, error_manager.WARNING)

  def _debug(self, row, column, message):
    error_manager.add(self.sheet_name, row + 1, column + 1, message, error_manager.DEBUG)

  def _traceback(self, message):
    package_logger.debug(message)

  def _format(self, row, column, message):
    if self.sheet_name == None:
      return f"{self.message}"
    elif row == None:
      return f"In sheet {self.sheet_name}: {message}"
    else:
      return f"In sheet {self.sheet_name} at [{row},{column}]: {message}"

  def _get_sheet_names(self, file_path):
    wb = load_workbook(file_path, read_only=True, keep_links=False)
    return wb.sheetnames

  def _sheet_present(self, file_path, sheet_name):
    sheet_names = self._get_sheet_names(file_path)
    return sheet_name in sheet_names

  def _check_cell_value(self, file_path, sheet_name, row, column, value):
    wb = load_workbook(file_path, read_only=True, keep_links=False)
    ws = wb[sheet_name]
    #print(f"CELL={ws.cell(row, column).value}")
    return str(ws.cell(row, column).value).upper() == value

  def _state_split(self, s):

    OUT = "out"
    IN_QUOTED = "in_quoted"
    OUT_QUOTED = "out_quoted"
    IN_NORMAL = "in_normal"
    ESC = "escape"

    state = OUT
    result = []
    current = []
    exit = ""
    for c in s:
      if state == OUT:
        current = []
        if c in ['"', "'"]:
          state = IN_QUOTED
          exit = c
        elif c.isspace():
          pass
        else:
          state = IN_NORMAL
          current.append(c)
      elif state == IN_QUOTED:
        if c == '\\':
          state = ESC
        elif c == exit:
          result.append("".join(current).strip())
          state = OUT_QUOTED
        else:
          current.append(c)
      elif state == OUT_QUOTED:
        if c == ',':
          state = OUT
        else:
          pass
      elif state == IN_NORMAL:
        if c == ',':
          result.append("".join(current).strip())
          state = OUT
        else:
          current.append(c)
      elif state == ESC:
        if c == exit:
          current.append(c)
          state = IN_QUOTED
        else:
          current.append('\\')
          current.append(c)
      else:
        raise BaseSheet.StateError

    if state == OUT or state == OUT_QUOTED:
      pass
    elif state == IN_NORMAL:
      result.append("".join(current).strip())
    else:
      raise BaseSheet.FormatError
    return result



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_model.narrative_content import NarrativeContent
import traceback
import re

class StudyDesignContentSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      self.items = []
      super().__init__(file_path=file_path, sheet_name='studyDesignContent', optional=True, converters={"sectionName": str})
      if self.success:
        current_level = 0
        new_level = 0
        current_parent = []
        previous_item = NarrativeContent(
          id=id_manager.build_id(NarrativeContent), 
          name="ROOT",
          sectionNumber="0",
          sectionTitle="Root",
          text="",
          contentChildIds=[]
        )
        self.items.append(previous_item)
        for index, row in self.sheet.iterrows():
          section_number = self.read_cell_by_name(index, 'sectionNumber')
          new_level = self._get_level(section_number)
          title = self.read_cell_by_name(index, 'sectionTitle')
          text = self.read_cell_by_name(index, 'text')
          name = self.read_cell_by_name(index, 'name')
          name = f"SECTION {section_number}" if name == "" else name
          #updated_text = text if self._standard_section(text) else self._wrap_div(text)
          #print(f"STD: Text='{updated_text}'")
          try:
            item = NarrativeContent(
              id=id_manager.build_id(NarrativeContent), 
              name=name,
              sectionNumber=section_number,
              sectionTitle=title,
              text=self._wrap_div(text),
              contentChildIds=[]
            )
          except Exception as e:  
            self._general_error(f"Failed to create Content object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          self.items.append(item)
          #print(f"ITEM: {item.id}")
          if new_level == current_level:
            # Same level
            parent = current_parent[-1]
            parent.contentChildIds.append(item.id)
          elif new_level > current_level:
            # Down
            if (new_level - current_level) > 1:
              self._error(index, self._get_column_index('sectionNumber'), f"Error with section number incresing by more than one level, section '{section_number}'.")
              raise BaseSheet.FormatError
            current_parent.append(previous_item)
            current_level = new_level
            parent = current_parent[-1]
            parent.contentChildIds.append(item.id)
          else:
            # Up
            for p_count in range(new_level, current_level):
              popped = current_parent.pop()
            parent = current_parent[-1]
            parent.contentChildIds.append(item.id)
            current_level = new_level
          previous_item = item
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _get_level(self, section_number):
    parts = section_number.split('.')
    return len(parts)

  # def _standard_section(self, text):
  #   result = re.match(r'SECTION\s*=', text.upper())
  #   #print(f"STD: {text} = {result}")
  #   return result

  def _wrap_div(self, text):
    return text if text.startswith("<div>") else f"<div>{text}</div>"


from usdm_excel.cdisc_ct_library import cdisc_ct_library
from usdm_excel.code_base import CodeBase

class CDISCCT(CodeBase):

  API_ROOT = 'https://api.library.cdisc.org/api'  

  def code(self, code, decode):
    return self._build(code=code, system=cdisc_ct_library.system, version=cdisc_ct_library.version, decode=decode)
 
  def code_for_attribute(self, klass, attribute, value):
    item = cdisc_ct_library.klass_and_attribute(klass, attribute, value)
    if item == None:
      return None
    return self._build(code=item['conceptId'], system=cdisc_ct_library.system, version=cdisc_ct_library.version, decode=item['preferredTerm'])



import pandas as pd

from src.usdm_excel.configuration_sheet import ConfigurationSheet
from src.usdm_excel.option_manager import Options, PrevNextOption, RootOption
from src.usdm_excel import ct_version_manager as ctvm
from src.usdm_excel import om

def test_defaults(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['Option X', 'Option 2', 'Option 3'], 'col_2': ['maybe', 'True', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("")
  assert om.get(Options.ROOT) == RootOption.API_COMPLIANT.value
  assert om.get(Options.PREVIOUS_NEXT) == PrevNextOption.NONE.value
  assert om.get(Options.DESCRIPTION) == ""

def test_set(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': ['Ct Version', 'SDR Prev Next', 'sdr ROOT', 'sdr Description'], 'col_2': ['THIS=that', 'sdr', 'SDR', 'No desc set']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  configuration = ConfigurationSheet("")
  assert om.get(Options.ROOT) == RootOption.API_COMPLIANT.value
  assert om.get(Options.PREVIOUS_NEXT) == PrevNextOption.NULL_STRING.value
  assert om.get(Options.DESCRIPTION) == "No desc set"
  assert ctvm.get('THIS') == 'that'
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "configuration"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "The SDR_ROOT option is now deprecated and will be ignored."



from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class StudyArm(ApiBaseModelWithIdNameLabelAndDesc):
  type: Code
  studyArmDataOriginDescription: str
  dataOriginType: Code



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_model.study_epoch import StudyEpoch
from usdm_model.study_arm import StudyArm
from usdm_model.study_element import StudyElement
from usdm_model.study_cell import StudyCell
from usdm_model.study_design import StudyDesign
from usdm_excel.alias import Alias
import pandas as pd
import traceback

class StudyDesignSheet(BaseSheet):

  VERSION_ROWS = { 
    'therapeuticAreas': {
      'NAME_ROW': 0,
      'DESCRIPTION_ROW': 1,
      'LABEL_ROW': -1,
      'TA_ROW': 2,
      'RATIONALE_ROW': 3,
      'BLINDING_ROW': 4,
      'INTENT_ROW': 5,
      'TYPES_ROW': 6,
      'INT_ROW': 7,
      'MAIN_TIMELINE_ROW': 8,
      'OTHER_TIMELINES_ROW': 9,
      'EPOCH_ARMS_START_ROW': 11
    },
    'label': { 
      'NAME_ROW': 0,
      'DESCRIPTION_ROW': 1,
      'LABEL_ROW': 2,
      'TA_ROW': 3,
      'RATIONALE_ROW': 4,
      'BLINDING_ROW': 5,
      'INTENT_ROW': 6,
      'TYPES_ROW': 7,
      'INT_ROW': 8,
      'MAIN_TIMELINE_ROW': 9,
      'OTHER_TIMELINES_ROW': 10,
      'EPOCH_ARMS_START_ROW': 12
    }
  }

  CHECK_ROW = 2
  PARAMS_NAME_COL = 0
  PARAMS_DATA_COL = 1

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesign', header=None)
      self.name = "TEST"
      self.description = "An Microsoft Excel test study design"
      self.label=""
      self.epochs = []
      self.epoch_names = {}
      self.arms = []
      self.arm_names = {}
      self.cells = []
      self.elements = {}
      self.study_designs = []
      self.therapeutic_areas = []
      self.rationale = ""
      self.blinding = None
      self.trial_intents = []
      self.trial_types = []
      self.intervention_model = None
      self.main_timeline = None
      self.other_timelines = []
      self.process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def process_sheet(self):
    key = self.read_cell(self.CHECK_ROW, self.PARAMS_NAME_COL)
    for rindex, row in self.sheet.iterrows():
      if rindex == self.VERSION_ROWS[key]['NAME_ROW']:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      if rindex == self.VERSION_ROWS[key]['LABEL_ROW']:
        self.label = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['DESCRIPTION_ROW']:
        self.description = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['TA_ROW']:
        self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['RATIONALE_ROW']:
        self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['BLINDING_ROW']:
        self.blinding = Alias().code(self.read_cdisc_klass_attribute_cell('StudyDesign', 'studyDesignBlindingScheme',rindex, self.PARAMS_DATA_COL), [])
      elif rindex == self.VERSION_ROWS[key]['INTENT_ROW']:
        self.trial_intents = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialIntentType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['TYPES_ROW']:
        self.trial_types = self.read_cdisc_klass_attribute_cell_multiple('StudyDesign', 'trialType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['INT_ROW']:
        self.intervention_model = self.read_cdisc_klass_attribute_cell('StudyDesign', 'interventionModel', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['MAIN_TIMELINE_ROW']:
        self.main_timeline = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROWS[key]['OTHER_TIMELINES_ROW']:
        self.other_timelines = self.read_cell_multiple(rindex, self.PARAMS_DATA_COL)
      else:
        pass

    start_row = self.VERSION_ROWS[key]['EPOCH_ARMS_START_ROW']
    resolved_epochs = [None] * self.sheet.shape[1] 
    resolved_arms = [None] * self.sheet.shape[0] 
    for rindex, row in self.sheet.iterrows():
      if rindex >= start_row:
        for cindex in range(0, len(self.sheet.columns)):
          epoch_index = cindex - 1
          cell = self.read_cell(rindex, cindex)
          if rindex == start_row:
            if cindex != 0:
              resolved_epochs[epoch_index] = self._add_epoch(cell)
          else:
            arm_index = rindex - start_row - 1
            if cindex == 0:
              resolved_arms[arm_index] = self._add_arm(cell)
            else:
              cell_elements = []
              element_names = self.read_cell_multiple(rindex, cindex)
              for name in element_names:
                element = self._add_element(name)
                if element is not None:
                  cell_elements.append(element.id)
              cell_arm = resolved_arms[arm_index].id
              cell_epoch = resolved_epochs[epoch_index].id
              if cell_arm is not None and cell_epoch is not None:
                self.cells.append(self._add_cell(arm=cell_arm, epoch=cell_epoch, elements=cell_elements))
              else:
                self._general_error(f"Cannot resolve arm and/or epoch for cell [{arm_index + 1},{epoch_index + 1}]")
              
    self.double_link(self.epochs, 'previousStudyEpochId', 'nextStudyEpochId')

    study_design = self._add_design(
      name=self.name,
      description=self.description,
      label=self.label,
      cells=self.cells,
      epochs=self.epochs,
      arms=self.arms,
      elements=list(self.elements.values()),
      intent_types=self.trial_intents, 
      trial_types=self.trial_types, 
      intervention_model=self.intervention_model,
      rationale=self.rationale, 
      blinding=self.blinding, 
      therapeutic_areas=self.therapeutic_areas
    )
    self.study_designs.append(study_design)

  def _add_arm(self, name):
    arm = cross_references.get(StudyArm, name)
    if arm is not None:
      if name not in self.arm_names:
        self.arm_names[name] = True
        self.arms.append(arm)
      return arm
    else:
      self._general_error(f"No arm definition found for arm '{name}'")
      return None

  def _add_epoch(self, name):
    epoch = cross_references.get(StudyEpoch, name)
    if epoch is not None:
      if name not in self.epoch_names:
        self.epoch_names[name] = True
        self.epochs.append(epoch)
      return epoch
    else:
      self._general_error(f"No epoch definition found for epoch '{name}'")
      return None
  
  def _add_element(self, name):
    element = cross_references.get(StudyElement, name)
    if element is not None:
      if name not in self.elements:
        self.elements[name] = element
      return element
    else:
      self._general_error(f"No element definition found for element '{name}'")
      return None

  def _add_cell(self, arm, epoch, elements):
    return StudyCell(
      id=id_manager.build_id(StudyCell), 
      studyArmId=arm,
      studyEpochId=epoch,
      studyElementIds=elements
    )

  def _add_design(self, name, description, label, epochs, arms, cells, elements, intent_types, trial_types, intervention_model, rationale, blinding, therapeutic_areas):
    return StudyDesign(
      id=id_manager.build_id(StudyDesign), 
      name=name,
      description=description,
      label=label,
      trialIntentTypes=intent_types,
      trialTypes=trial_types,
      interventionModel=intervention_model,
      studyCells=cells,
      studyArms=arms,
      studyEpochs=epochs,
      studyElements=elements,
      therapeuticAreas=therapeutic_areas,
      studyDesignRationale=rationale,
      studyDesignBlindingScheme=blinding,
      contents=[]
    )



from typing import Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .syntax_template_dictionary import SyntaxTemplateDictionary

class SyntaxTemplate(ApiBaseModelWithIdNameLabelAndDesc):
  instanceType: Literal['ENDPOINT', 'OBJECTIVE', 'ELIGIBILITY_CRITERIA']
  text: str = None
  dictionaryId: Union[str, None] = None



from usdm_excel.id_manager import id_manager
from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_excel.study_soa_v2_sheet.activity import Activity
import pandas as pd

class Activities():
  
  def __init__(self, parent):
    self.parent = parent
    self.items = []
    self.map = {}
    for row_index, col_def in parent.sheet.iterrows():
      if row_index >= SoAColumnRows.FIRST_ACTIVITY_ROW:
        activity = Activity(self.parent, row_index)
        self.items.append(activity)



import json

class ExportAsYworksDict():

  FULL = "full"
  TIMELINE = "timeline"

  def __init__(self, study, view=FULL):
    self.study = study
    self.nodes = []
    self.edges = []
    self.add_edges = []
    self.node_index = 1
    self.edge_index = 1
    self.id_node_index_map = {}
    self.edge_attributes = {
      'ScheduledActivityInstance': [
        'timepointActivityIds',
        'scheduledActivityInstanceEncounterId',
        'activityIds',
        'epochId',
        'defaultConditionId',
      ],
      'ScheduledDecisionInstance': [
        'timepointActivityIds',
        'epochId',
        'defaultConditionId',
        'conditionAssignments',
      ],
      'Activity': [
        'bcSurrogateIds',
        'bcCategoryIds',
        'biomedicalConceptIds',
      ],
      'Timing': [
        'relativeFromScheduledInstanceId',
        'relativeToScheduledInstanceId',
      ],
      'Estimand': [
        'treatment',
        'variableOfInterest',
      ],
      'ScheduledTimeline': [
        'activityTimelineId',
      ],
      'StudyCell': [
        'studyArmId',
        'studyEpochId',
        'studyElementIds',
      ]
    }
    # self.fix_id_name = {
    #   'scheduledActivityInstanceId': 'scheduledInstanceId',
    #   'scheduledDecisionInstanceId': 'scheduledInstanceId',
    #   'biomedicalConceptSurrogateId': 'bcSurrogateId',
    #   'biomedicalConceptPropertyId': 'bcPropertyId'
    # }
    self.order_attributes = []
    self.ignore_klass = []
    self.collapse_klass = []
    if view == self.__class__.TIMELINE:
      self.ignore_klass = [
        'StudyProtocolVersion', 'StudyIdentifier', 'Indication', 
        'InvestigationalIntervention', 'Objective', 'StudyDesignPopulation', 
        'Estimand', 'StudyCell', 'TransitionRule', 'StudyArm', 'StudyEpoch', 'StudyElement',
        'BiomedicalConcept', 'BiomedicalConceptCategory', 'BiomedicalConceptSurrogate', 
        'Procedure', 'AliasCode', 'NarrativeContent', 'StudyAmendment', 'EligibilityCriteria', 
        'StudyProtocolDocument', 'GovernanceDate'
      ]
      self.collapse_klass = ['Code']
      #self.order_attributes = [
      #  'scheduleSequenceNumber'
      #]
    #self.sequence_number_map = {}
      
  def export(self):
    node = json.loads(self.study.to_json_with_type())
    #print(f"NODE: {node}")
    self._process_node(node)
    for edge in self.add_edges:
      if edge['end'] in self.id_node_index_map:
        edge['id'] = self.edge_index
        edge['end'] = self.id_node_index_map[edge['end']]
        self.edges.append(edge)
        self.edge_index += 1
      else:
        print("***** %s -edge-> %s [%s] *****" % (edge['start'], edge['end'], edge['properties']))
    return self.nodes, self.edges
  
  def _process_node(self, node):
    if type(node) == list:
      result = []
      for item in node:
        indexes = self._process_node(item)
        if indexes is None:
          return None
        result = result + indexes
      return result
    elif type(node) == dict:
      if node == {}:
        return []
      properties = {}
      id_field, klass = self._get_id_field_and_klass(node)
      if not klass:
        return None
      if klass in self.ignore_klass:
        return None
      if node[id_field] in self.id_node_index_map:
        return [self.id_node_index_map[node[id_field]]]
      this_node_index = self.node_index
      self.node_index += 1
      if klass in self.collapse_klass:
        return []
      for key, value in node.items():
        # Special case, get the ids for the sequence numbers but within scope of each timeline
        # if key == "scheduleTimelineInstances":
        #   self.sequence_number_map = {}
        #   for item in value:
        #     self.sequence_number_map[item['scheduleSequenceNumber']] = item['scheduledInstanceId']
        # Link the sequence numbers
        # if key in self.order_attributes:
        #   seq = value + 1
        #   if seq in self.sequence_number_map:
        #     self.add_edges.append( { 'start': this_node_index, 'end': self.sequence_number_map[seq], 'properties': { 'label': key, 'type': 'Order' }})
        if klass in self.edge_attributes and key in self.edge_attributes[klass]:
          if key == "conditionAssignments":
            # Special case, array of arrays of condition and link id
            for item in value:
              self.add_edges.append( { 'start': this_node_index, 'end': item[1], 'properties': { 'label': key, 'type': 'Condition' }})
          # elif key == "parameterMap":
          #   properties[key] = value
          elif type(value) == list:
            for item in value:
              if item:
                self.add_edges.append( { 'start': this_node_index, 'end': item, 'properties': { 'label': key, 'type': 'List' }})
          else:
            if value:
              self.add_edges.append( { 'start': this_node_index, 'end': value, 'properties': { 'label': key, 'type': 'Other' }})
        else:
          indexes = self._process_node(value)
          if indexes == None:
            properties[key] = {}
          elif indexes == []:
            properties[key] = value
          else:
            for index in indexes:
              self.edges.append( {'id': self.edge_index, 'start': this_node_index, 'end': index, 'properties': {'label': key}})
              self.edge_index += 1
      if klass == "Study":
        node[id_field] = "Study"
      properties['label'] = node[id_field]
      self.nodes.append({ 'id': this_node_index, 'properties': properties })
      self.id_node_index_map[properties[id_field]] = this_node_index
      return [this_node_index]
    else:
      return []
    
  def _get_id_field_and_klass(self, node):
    try:
      klass = node['_type']
    # id_name = "%s%s" % (stringcase.camelcase(klass), "Id")
    # if id_name in self.fix_id_name:
    #   id_name = self.fix_id_name[id_name]
      return 'id', klass
    except Exception as e:
      #print(f"NODE: {node}")
      return 'id', None




from typing import List, Union
from .syntax_template import SyntaxTemplate
from .code import Code
from .endpoint import Endpoint

class Objective(SyntaxTemplate):
  level: Union[Code, None] = None
  objectiveEndpoints: List[Endpoint] = []


from typing import List, Union
from .syntax_template import SyntaxTemplate
from .code import Code

class EligibilityCriteria(SyntaxTemplate):
  category: Code
  identifier: str
  nextCriterionId: Union[str, None] = None
  previousCriterionId: Union[str, None] = None


import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.base_sheet import BaseSheet
from usdm_model.code import Code

# def test_sheet_cell_value(mocker):
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', 'd']}
#   mock_read = mocker.patch("openpyxl.load_workbook")
#   mock_read.return_value = pd.DataFrame.from_dict(data)
#   test_data = [
#     ({'row': 0, 'column': 0, 'value': 4}, False),
#     ({'row': 0, 'column': 1, 'value': 'e'}, False),
#     ({'row': 0, 'column': 0, 'value': 3}, True),
#     ({'row': 0, 'column': 1, 'value': 'a'}, True),
#   ]
#   for test in test_data:
#     base = BaseSheet("", "sheet", require=test[0])
#     if not test[1]:
#       assert(base) == None
#     else:
#       assert(base) != None

def test_cell_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', 'c', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0,0,False),
    (3,0,False),
    (0,1,True),
    (3,1,True)
  ]
  for test in test_data:
    assert(base.cell_empty(test[0],test[1])) == test[2]

def test_cell_empty_legacy(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', '-', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0,0,'3',False),
    (3,0,'0',False),
    (0,1,'',True),
    (2,1,'',True),
    (3,1,'',True)
  ]
  for test in test_data:
    value, is_null = base.read_cell_empty_legacy(test[0],test[1])
    assert(value) == test[2]
    assert(is_null) == test[3]

def test_read_cell_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', '-', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0,0,'3','3'),
    (3,0,'0','0'),
    (0,1,'',''),
    (2,1,'','-'),
    (3,1,'','')
  ]
  for test in test_data:
    assert(base.read_cell_empty(test[0],test[1],'-')) == test[2]
    assert(base.read_cell_empty(test[0],test[1],'=')) == test[3]

def test_read_cell(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0,0,'3'),
    (3,0,'0'),
    (0,1,'a'),
    (3,1,'')
  ]
  for test in test_data:
    assert(base.read_cell(test[0],test[1])) == test[2]

def test_read_cell_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  assert(base.read_cell(6,8)) == ''
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 7
  assert mock_error.call_args[0][2] == 9
  assert mock_error.call_args[0][3] == "Error (index 8 is out of bounds for axis 0 with size 2) reading cell"
  
def test_read_cell_by_name(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0,'Name','tom'),
    (2,'Name','juli'),
    (0,'Age','10'),
    (2,'Age','14')
  ]
  for test in test_data:
    assert(base.read_cell_by_name(test[0],test[1])) == test[2]

def test_read_cell_by_name_default(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", "sheet")
  assert(base.read_cell_by_name(0, 'NameX', 'Default')) == 'Default'
  assert(base.read_cell_by_name(0, 'Name', 'Default')) == 'tom'
  assert(base.read_cell_by_name(0, ['NameX', 'XXX'], 'Default')) == 'Default'
  assert(base.read_cell_by_name(0, ['NameX', 'XXX', 'Age'], 'Default')) == '10'

def test_read_cell_by_name_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", "sheet")
  assert(base.read_cell_by_name(1,'Not There')) == ''
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 2
  assert mock_error.call_args[0][2] == -1
  assert mock_error.call_args[0][3] == "Error reading cell 'Not There'"
  
def test_read_cell_multiple(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['tom', ''], 
    ['nick', 'Sam'], 
    ['juli', ' Fred, Dick,   Harry  '], 
    ['andy', 'John  , Jane'], 
    ['andy', '"John, & Fred", Jane'],
    ['andy', '"John, \\" & Fred", Jane']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0,1,[]),
    (2,1,['Fred', 'Dick', 'Harry']),
    (3,1,['John', 'Jane']),
    (4,1,['John, & Fred', 'Jane']),
    (5,1,['John, " & Fred', 'Jane'])
  ]
  for test in test_data:
    assert(base.read_cell_multiple(test[0],test[1])) == test[2]

# read_cell_with_previous

def test_read_boolean_cell_by_name(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  #data = [[0, 1, 2, 3, 4, 5, 6, 7, 8], ['a', 'y', 'Y', 'true', 'True', 'yes', 1, '1', '']]
  data = [[0, 'a'], [1, 'y'], [2, 'Y'], [3, 'true'], [4, 'True'], [5, 'yes'], [6, 1,], [7, '1'], [8, ''], [9, 'T']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0,'Children',False),
    (1,'Children',True),
    (2,'Children',True),
    (3,'Children',True),
    (4,'Children',True),
    (5,'Children',True),
    (6,'Children',True),
    (7,'Children',True),
    (8,'Children',False),
    (9,'Children',True),
  ]
  for test in test_data:
    assert(base.read_boolean_cell_by_name(test[0],test[1])) == test[2]

def test_read_description_by_name(mocker):
  mock_option = mocker.patch("usdm_excel.om.get")
  mock_option.side_effect=['xxx', 'xxx', 'xxx']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[0, ''], [1, 'something'], [2, '  ']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0,'Children','xxx'),
    (1,'Children','something'),
    (2,'Children','xxx'),
  ]
  for test in test_data:
    assert(base.read_description_by_name(test[0],test[1])) == test[2]

@xfail
def test_read_cell_with_previous():
  assert 0

@xfail
def test_read_other_code_cell_by_name():
  assert 0

@xfail
def test_read_other_code_cell():
  assert 0

@xfail
def test_read_other_code_cell_multiple_by_name():
  assert 0

@xfail
def test_read_other_code_cell_mutiple():
  assert 0

def test_column_present(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  assert(base.column_present("Name")) == 0
  assert(base.column_present(["Name"])) == 0
  assert(base.column_present(["Children"])) == 1
  assert(base.column_present(["ChildrenX", "Children"])) == 1
  with pytest.raises(BaseSheet.FormatError):
    assert(base.column_present(["Fred"]))
  
def test_read_cdisc_klass_attribute_cell_by_name(mocker):
  expected = Code(id='CodeX', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected, None]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[0, 'a'], [1, ''], [2, 'c']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0, 'Children', expected, ""),
    (1, 'Children', None, "sheet", 2, 2, "Empty cell detected where CDISC CT value expected."),
    (2, 'Children', None, "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_by_name( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test_read_cdisc_klass_attribute_cell(mocker):
  expected = Code(id='CodeX', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected, None]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', '', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0, 1, expected, ""),
    (1, 1, None, "sheet", 2, 2, "Empty cell detected where CDISC CT value expected."),
    (2, 1, None, "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]
     
@xfail
def test_read_cdisc_klass_attribute_cell_multiple_by_name(mocker):
  assert 0
  
def test_read_cdisc_klass_attribute_cell_multiple(mocker):
  expected_1 = Code(id='CodeX1', code='code1', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='CodeX2', code='code2', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, None]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a,b', '', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", "sheet")
  test_data = [
    (0, 1, [expected_1, expected_2], ""),
    (1, 1, [], "sheet", 2, 2, "Empty cell detected where multiple CDISC CT values expected."),
    (2, 1, [], "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_multiple( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test_read_cdisc_klass_attribute_cell_multiple_by_name(mocker):
  expected_1 = Code(id='CodeX1', code='code1', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='CodeX2', code='code2', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, None]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[0, 'a,b'], [1, ''], [2, 'c']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", "sheet")
  test_data = [
    (0, 'Children', [expected_1, expected_2], ""),
    (1, 'Children', [], "sheet", 2, 2, "Empty cell detected where multiple CDISC CT values expected."),
    (2, 'Children', [], "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_multiple_by_name( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test__decode_other_cell(mocker):
  expected = Code(id='Code_1', code='c', codeSystem='a', codeSystemVersion='3', decode="d")
  mock_version = mocker.patch("usdm_excel.ct_version_manager.get")
  mock_version.side_effect=['3']
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Code_1']
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", "sheet")
  test_data = [
    ("", 0, 0, None, ""),
    ("xxx", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'xxx', no ':' detected"),
    ("a:", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'a:', no '=' detected"),
    ("a:c", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'a:c', no '=' detected"),
    ("a:c=d", 1, 1, expected, "")
  ]
  for test in test_data:
    assert(base._decode_other_code(test[0], test[1], test[2])) == test[3]
    if not test[4] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[4]
      assert mock_error.call_args[0][1] == test[5]
      assert mock_error.call_args[0][2] == test[6]
      assert mock_error.call_args[0][3] == test[7]

def test__state_split(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", "sheet")
  test_data = [
    ('111', ['111']),
    ('111,    ', ['111']),
    ('"111", 222, 333', ['111', '222', '333']),
    ('"111", 222, 333, "4"', ['111', '222', '333', '4']),
    ('"111 \\" quote", 222, 333', ['111 " quote', '222', '333'])
  ]
  for test in test_data:
    assert(base._state_split(test[0])) == test[1]

  with pytest.raises(BaseSheet.FormatError):
    base._state_split('123, "456')

def test__to_int(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", "sheet")
  assert(base._to_int("4")) == 4
  assert(base._to_int("dd")) == None


from pydantic import BaseModel
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_identifiers_sheet.study_identifiers_sheet import StudyIdentifiersSheet
from usdm_excel.study_design_sheet.study_design_sheet import StudyDesignSheet
from usdm_excel.study_soa_sheet.study_soa_sheet import StudySoASheet
from usdm_excel.study_soa_v2_sheet.study_soa_v2_sheet import StudySoAV2Sheet
from usdm_excel.study_design_ii_sheet.study_design_ii_sheet import StudyDesignIISheet
from usdm_excel.study_design_population_sheet.study_design_population_sheet import StudyDesignPopulationSheet
from usdm_excel.study_design_objective_endpoint_sheet.study_design_objective_endpoint_sheet import StudyDesignObjectiveEndpointSheet
from usdm_excel.study_design_estimands_sheet.study_design_estimands_sheet import StudyDesignEstimandsSheet
from usdm_excel.study_design_procedure_sheet.study_design_procedure_sheet import StudyDesignProcedureSheet
from usdm_excel.study_design_encounter_sheet.study_design_encounter_sheet import StudyDesignEncounterSheet
from usdm_excel.study_design_element_sheet.study_design_element_sheet import StudyDesignElementSheet
from usdm_excel.study_design_arm_sheet.study_design_arm_sheet import StudyDesignArmSheet
from usdm_excel.study_design_epoch_sheet.study_design_epoch_sheet import StudyDesignEpochSheet
from usdm_excel.study_design_activity_sheet.study_design_activity_sheet import StudyDesignActivitySheet
from usdm_excel.study_design_timing_sheet.study_design_timing_sheet import StudyDesignTimingSheet
from usdm_excel.study_design_content_sheet.study_design_content_sheet import StudyDesignContentSheet
from usdm_excel.study_design_amendment_sheet.study_design_amendment_sheet import StudyDesignAmendmentSheet
from usdm_excel.study_design_dictionary_sheet.study_design_dictionary_sheet import StudyDesignDictionarySheet
from usdm_excel.study_design_eligibility_criteria_sheet.study_design_eligibility_criteria_sheet import StudyDesignEligibilityCriteriaSheet
from usdm_excel.alias import Alias
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_excel.option_manager import *
from usdm_model.study import Study
from usdm_model.study_version import StudyVersion
from usdm_model.study_protocol_document_version import StudyProtocolDocumentVersion
from usdm_model.study_protocol_document import StudyProtocolDocument
from usdm_model.wrapper import Wrapper
from usdm_model.governance_date import GovernanceDate
from usdm_model.geographic_scope import GeographicScope
from usdm_excel.narrative_content import NarrativeContent
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.iso_3166 import ISO3166
import traceback
import datetime

class StudySheet(BaseSheet):

  NAME_ROW = 0
  TITLE_ROW = 1
  VERSION_ROW = 2
  TYPE_ROW = 3
  PHASE_ROW = 4
  ACRONYM_ROW = 5
  RATIONALE_ROW = 6
  TA_ROW = 7
  BRIEF_TITLE_ROW = 8
  OFFICAL_TITLE_ROW = 9
  PUBLIC_TITLE_ROW = 10
  SCIENTIFIC_TITLE_ROW = 11
  PROTOCOL_VERSION_ROW = 12
  PROTOCOL_STATUS_ROW = 13

  DATES_HEADER_ROW = 15
  DATES_DATA_ROW = 16
  
  PARAMS_DATA_COL = 1

  STUDY_VERSION_DATE = 'study_version'
  PROTOCOL_VERSION_DATE = 'protocol_document'

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='study', header=None)
      self.date_categories = [self.STUDY_VERSION_DATE, self.PROTOCOL_VERSION_DATE]
      self.soa_version = None
      self.phase = None
      self.version = None
      self.type = None
      self.name = None
      self.brief_title = None
      self.official_title = None
      self.public_title = None
      self.scientific_title = None
      self.protocol_version = None
      self.protocol_status = None
      self.title = None
      self.acronym = None
      self.rationale = None
      self.study = None
      self.study_version = None
      self.protocol_document_version = None
      self.therapeutic_areas = []
      self.timelines = {}
      self.dates = {}
      for category in self.date_categories:
        self.dates[category] = []

      # Process all the sheets
      self._process_sheet()
      self.study_amendments = StudyDesignAmendmentSheet(file_path)
      self.study_identifiers = StudyIdentifiersSheet(file_path)
      self.procedures = StudyDesignProcedureSheet(file_path)
      self.encounters = StudyDesignEncounterSheet(file_path)
      self.elements = StudyDesignElementSheet(file_path)
      self.arms = StudyDesignArmSheet(file_path)
      self.epochs = StudyDesignEpochSheet(file_path)
      self.timings = StudyDesignTimingSheet(file_path)
      self.activities = StudyDesignActivitySheet(file_path)
      self.study_design = StudyDesignSheet(file_path)
      self._process_soa(file_path)
      self.ii = StudyDesignIISheet(file_path)
      self.study_populations = StudyDesignPopulationSheet(file_path)
      self.contents = StudyDesignContentSheet(file_path)
      self.dictionaries = StudyDesignDictionarySheet(file_path)
      self.oe = StudyDesignObjectiveEndpointSheet(file_path)
      self.eligibility_criteria = StudyDesignEligibilityCriteriaSheet(file_path)
      self.estimands = StudyDesignEstimandsSheet(file_path)

      # Study Design assembly
      study_design = self.study_design.study_designs[0]
      study_design.studyScheduleTimelines.append(self.soa.timeline)
      study_design.encounters = self.encounters.items
      study_design.activities = self.soa.activities
      activity_ids = [item.id for item in study_design.activities]
      study_design.biomedicalConcepts = self.soa.biomedical_concepts
      study_design.bcSurrogates = self.soa.biomedical_concept_surrogates
      for key,tl in self.timelines.items():
        study_design.studyScheduleTimelines.append(tl.timeline)
        for activity in tl.activities:
          if activity.id not in activity_ids:
            study_design.activities.append(activity)
            activity_ids.append(activity.id)
        study_design.biomedicalConcepts += tl.biomedical_concepts
        study_design.bcSurrogates += tl.biomedical_concept_surrogates
      study_design.studyIndications = self.ii.indications
      study_design.studyInvestigationalInterventions = self.ii.interventions
      study_design.studyPopulations = self.study_populations.populations
      study_design.studyObjectives = self.oe.objectives
      study_design.studyEstimands = self.estimands.estimands
      study_design.studyEligibilityCritieria = self.eligibility_criteria.items
      study_design.dictionaries = self.dictionaries.items
      
      # Final assembly
      try:
        self.protocol_document_version = StudyProtocolDocumentVersion(
          id=id_manager.build_id(StudyProtocolDocumentVersion), 
          briefTitle=self.brief_title,
          officialTitle=self.official_title,
          publicTitle=self.public_title,
          scientificTitle=self.scientific_title,
          protocolVersion=self.protocol_version,
          protocolStatus=self.protocol_status,
          dateValues=self.dates[self.PROTOCOL_VERSION_DATE]
          )
        self.protocol_document_version.contents = self.contents.items
        cross_references.add(self.protocol_document_version.id, self.protocol_document_version)
      except Exception as e:
        self._general_error(f"Failed to create StudyProtocolDocumentVersion object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")

      try:
        study_protocol_document = StudyProtocolDocument(
          id=id_manager.build_id(StudyProtocolDocument), 
          name=f"Protocol_Document_{self.name}", 
          versions=[self.protocol_document_version])
      except Exception as e:
        self._general_error(f"Failed to create StudyProtocolDocument object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")

      try:
        #f"DATES SV: {self.dates[self.STUDY_VERSION_DATE]}")
        self.study_version = StudyVersion(
          id=id_manager.build_id(StudyVersion),
          studyTitle=self.title,
          studyVersion=self.version,
          type=self.type,
          studyPhase=self.phase,
          businessTherapeuticAreas=self.therapeutic_areas,
          studyRationale=self.rationale,
          studyAcronym=self.acronym,
          studyIdentifiers=self.study_identifiers.identifiers,
          documentVersionId=self.protocol_document_version.id,
          studyDesigns=self.study_design.study_designs,
          dateValues=self.dates[self.STUDY_VERSION_DATE],
          amendments=self.study_amendments.items
        )
        cross_references.add(self.study_version.id, self.study_version)
      except Exception as e:
        self._general_error(f"Failed to create StudyVersion object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")

      try:
        self.study = Study(
          id=None, # No Id, will be allocated a UUID
          name=f"Study_{self.name}", 
          versions=[self.study_version],
          documentedBy=study_protocol_document
        )
        cross_references.add("STUDY", self.study)
      except Exception as e:
        self._general_error(f"Failed to create Study object, exception {e}")
        self._traceback(f"{traceback.format_exc()}")

    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  # def study_sponsor(self):
  #   return self.cdisc_code(code="C93453", decode="Study Registry")

  # def study_regulatory(self):
  #   return self.cdisc_code(code="C93453", decode="Study Registry")

  def the_study(self):
    return self.study
  
  def api_root(self):
    return Wrapper(study=self.study)

  def to_html(self):
    return NarrativeContent(self.title, self.study).to_html()

  def to_pdf(self):
    return NarrativeContent(self.title, self.study).to_pdf()

  def _process_sheet(self):
    fields = ['category', 'name', 'description', 'label', 'type', 'date', 'scopes']    
    for rindex, row in self.sheet.iterrows():
      if rindex == self.NAME_ROW:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TITLE_ROW:
        self.title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.VERSION_ROW:
        self.version = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TYPE_ROW:
        self.type = self.read_cdisc_klass_attribute_cell('Study', 'studyType', rindex, self.PARAMS_DATA_COL)
      elif rindex == self.PHASE_ROW:
        phase = self.read_cdisc_klass_attribute_cell('Study', 'studyPhase', rindex, self.PARAMS_DATA_COL)
        self.phase = Alias().code(phase, [])
      elif rindex == self.ACRONYM_ROW:
        self.acronym = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.RATIONALE_ROW:
        self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.TA_ROW:
        self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.BRIEF_TITLE_ROW:
        self.brief_title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.OFFICAL_TITLE_ROW:
        self.official_title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.PUBLIC_TITLE_ROW:
        self.public_title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.SCIENTIFIC_TITLE_ROW:
        self.scientific_title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.PROTOCOL_VERSION_ROW:
        self.protocol_version = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif rindex == self.PROTOCOL_STATUS_ROW:
        self.protocol_status = self.read_cdisc_klass_attribute_cell('StudyProtocolVersion', 'protocolStatus', rindex, self.PARAMS_DATA_COL) 
      elif rindex >= self.DATES_DATA_ROW:
        record = {}
        for cindex in range(0, len(self.sheet.columns)):
          field = fields[cindex]
          if field == 'category':
            cell = self.read_cell(rindex, cindex)
            if cell.lower() in self.date_categories:
              category = cell.lower()
            else:
              categories = ', '.join(f'"{w}"' for w in self.date_categories)
              self._error(rindex, cindex, f"Date category not recognized, should be one of {categories}, defaults to '{self.date_categories[0]}'")
              category = self.date_categories[0]
          elif field == 'type':
            record[field] = self.read_cdisc_klass_attribute_cell('GovernanceDate', 'type', rindex, cindex)
            #print(f"TYPE: {record[field]}") 
          elif field == 'date':
            cell = self.read_cell(rindex, cindex)
            record[field] = datetime.datetime.strptime(cell, '%Y-%m-%d %H:%M:%S')
          elif field == 'scopes':
            record[field] = self._read_scope_cell(rindex, cindex)
          else:
            cell = self.read_cell(rindex, cindex)
            record[field] = cell
        try:
          scopes = []
          for scope in record['scopes']:
            scope = GeographicScope(
              id=id_manager.build_id(GeographicScope), 
              type=scope['type'], 
              code=scope['code']
            )
            scopes.append(scope)
            #print(f"SCOPE: {scope}")
        except Exception as e:
          self._general_error(f"Failed to create GeographicScope object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
          #print(f"SCOPE: {traceback.format_exc()}")
        try:
          date = GovernanceDate(
            id=id_manager.build_id(GovernanceDate),
            name=record['name'],
            label=record['label'],
            description=record['description'],
            type=record['type'],
            dataValue=record['date'],
            geographicScopes=scopes
          )
          self.dates[category].append(date)
          #print(f"DATE: {date}")
        except Exception as e:
          self._general_error(f"Failed to create GovernanceDate object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
          #print(f"DATE: {traceback.format_exc()}")

  def _process_soa(self, file_path):
    for timeline in self.study_design.other_timelines:
      tl = self._process_timeline(file_path, timeline)
      self.timelines[timeline] = tl
      cross_references.add(timeline, tl.timeline)
    self.soa = self._process_timeline(file_path, self.study_design.main_timeline, True)

  def _process_timeline(self, file_path, timeline, main_timeline=False):
    if not self.soa_version:
      self._general_info("Detecting SoA sheet version ...")
      tl = StudySoASheet(file_path, timeline, main=main_timeline, require={'row': 1, 'column': 3, 'value': 'EPOCH'})
      if tl.success:
        self._general_info("SoA sheet version 1 detected")
        self.soa_version = 1 
      else:
        self._general_info("SoA sheet version 2 detected")
        self.soa_version = 2
        tl = StudySoAV2Sheet(file_path, timeline, main=main_timeline)
        #print("---- SoA Timing ----")
        tl.set_timing_references(self.timings.items)
    elif self.soa_version == 1:
      self._general_info("Set to SoA sheet version 1")
      tl = StudySoASheet(file_path, timeline, main=main_timeline)
    else:
      self._general_info("Set to SoA sheet version 2")
      tl = StudySoAV2Sheet(file_path, timeline, main=main_timeline)
      #print("---- SoA Timing ----")
      tl.set_timing_references(self.timings.items)
    return tl

  def _read_scope_cell(self, row_index, col_index):
    result = []
    value = self.read_cell(row_index, col_index)
    if value.strip() == "":
      self._error(row_index, col_index, "Empty cell detected where multiple geographic scope CT values expected")
      return result
    else:
      for item in self._state_split(value):
        #print(f"SCOPE ITEM: {item}")
        if item.upper().strip() == "GLOBAL":
          # If we ever find global just return the one code
          return [{'type': CDISCCT().code_for_attribute('GeographicScope', 'type', 'Global'), 'code': None}]
        else: 
          code = None
          if item.strip():
            outer_parts = item.split(":")
            if len(outer_parts) == 2:
              system = outer_parts[0].strip()
              value = outer_parts[1].strip()
              if system.upper() == "REGION":
                pt = 'Region'
                code = ISO3166().region_code(value)
              elif system.upper() == "COUNTRY":
                pt = 'Country'
                code = ISO3166().code(value)
              else:
                self._error(row_index, col_index, f"Failed to decode geographic scope data {outer_parts}, must be either Global, Region using UN M49 codes, or Country using ISO3166 codes")
            else:
              self._error(row_index, col_index, f"Failed to decode geographic scope data {outer_parts}, no ':' detected")
          else:
            self._error(row_index, col_index, f"Failed to decode geographic scope data {item}, appears empty")
          if code:
            result.append({'type': CDISCCT().code_for_attribute('GeographicScope', 'type', pt), 'code':  Alias().code(code, [])})
      return result



from typing import List, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode

class BiomedicalConceptCategory(ApiBaseModelWithIdNameLabelAndDesc):
  bcCategoryChildIds: List[str] = []
  bcCategoryMemberIds: List[str] = []
  code: Union[AliasCode, None] = None



from enum import Enum

class Options(Enum):
  PREVIOUS_NEXT = 'previous'
  ROOT = 'root'
  DESCRIPTION = 'description'

class PrevNextOption(Enum):
  NULL_STRING = 'null_string'
  NONE = 'none'

class RootOption(Enum):
  SDR_COMPATABLE = 'sdr_compatable'
  API_COMPLIANT = 'api_compliant'

class OptionManager():

  def __init__(self):
    self.items = {}

  def clear(self):
    self.items = {}

  def set(self, name, value):
    name = self._to_string(name)
    value = self._to_string(value)
    self.items[name] = value

  def get(self, name):
    name = self._to_string(name)
    if name in self.items:
      return self.items[name]
    else:
      return ""
  
  def _to_string(self, item):
    if isinstance(item, Enum):
      return item.value
    else:
      return str(item)
    
option_manager = OptionManager()




from yattag import Doc
from bs4 import BeautifulSoup   
from usdm_excel.cross_ref import cross_references
from usdm_excel.logger import logging
import docraptor
import re

class NarrativeContent():

  class LogicError(Exception):
    pass

  def __init__(self, doc_title, study):
    self.study = study
    self.study_version = study.versions[0]
    self.study_design = self.study_version.studyDesigns[0]
    self.protocol_document_version = self.study.documentedBy.versions[0]
    self.doc_title = doc_title
    if self.protocol_document_version.id != self.study.versions[0].documentVersionId:
      logging.error(f"Failed to initialise NarrativeContent for document creation, ids did not match")
      raise self.LogicError(f"Failed to initialise NarrativeContent for document creation, ids did not match")

  def to_pdf(self):
    doc_api = docraptor.DocApi()
    doc_api.api_client.configuration.username = 'YOUR_API_KEY_HERE'

    document_content = self.to_html()
    try:
      response = doc_api.create_doc({
        'test': True,  # test documents are free but watermarked
        'document_type': 'pdf',
        'document_content': document_content,
        # 'document_url': 'https://docraptor.com/examples/invoice.html',
        # 'javascript': True,
        # 'prince_options': # {
        #    'media': 'print', # @media 'screen' or 'print' CSS
        #    'baseurl': 'https://yoursite.com', # the base URL for any relative URLs
        # },
      })
      binary_formatted_response = bytearray(response)
      return binary_formatted_response
    except docraptor.rest.ApiException as error:
      pass
      #print(error.status)
      #print(error.reason)
      #print(error.body)

  def to_html(self):
    root = self.protocol_document_version.contents[0]
    doc = Doc()
    doc.asis('<!DOCTYPE html>')
    style = """
      /* Create a running element */
      #header-and-footer {
        position: running(header-and-footer);
        text-align: right;
      }

      /* Add that running element to the top and bottom of every page */
      @page {
        @top {
          content: element(header-and-footer);
        }
        @bottom {
          content: element(header-and-footer);
        }
      }

      /* Add a page number */
      #page-number {
        content: "Page " counter(page);
      }

      /* Create a title page with a full-bleed background and no header */
      #title-page {
        page: title-page;
      }

      @page title-page {
        @top {
          content: "";
        }
      }

      #title-page h1 {
        padding: 200px 0 40px 0;
        font-size: 30px;
      }

      /* Dynamically create a table of contents with leaders */
      #table-of-contents a {
        content: target-content(attr(href)) leader('.') target-counter(attr(href), page);
        color: #000000;
        text-decoration: none;
        display: block;
        padding-top: 5px;
      }

      /* Float the footnote to a footnotes area on the page */
      .footnote {
        float: footnote;
        font-size: small;
      }

      .page {
        page-break-after: always;
      }

      body {
        counter-reset: chapter;
        font-family: 'Times New Roman';
        color: #000000;
      }
    </style>
    <link href='https://fonts.googleapis.com/css2?family=Open+Sans:wght@400;600;800&display=swap' rel='stylesheet'>
    """
    chapters = []
    for id in root.contentChildIds:
      content = next((x for x in self.protocol_document_version.contents if x.id == id), None)
      level = len(content.sectionNumber.split('.'))
      if level == 1:
        chapters.append(f'<a href="#section-{content.sectionNumber}"></a>')
    front_sheet = f"""
      <div id="title-page" class="page">
        <h1>{self.doc_title}</h1>
        <div id="header-and-footer">
          <span id="page-number"></span>
        </div>
      </div>
      <div id="toc-page" class="page">
        <div id="table-of-contents">
          {''.join(chapters)}
        </div>
        <div id="header-and-footer">
          <span id="page-number"></span>
        </div>
      </div>
    """
    with doc.tag('html'):
      with doc.tag('head'):
        with doc.tag('style'):
          doc.asis(style)      
      with doc.tag('body'):
        doc.asis(front_sheet)    
        for id in root.contentChildIds:
          content = next((x for x in self.protocol_document_version.contents if x.id == id), None)
          if content:
            self._content_to_html(content, doc)
    return doc.getvalue()
  
  def _content_to_html(self, content, doc):
    level = len(content.sectionNumber.split('.'))
    klass = "page" if level == 1 else ""
    id = f"section-{content.sectionNumber}"
    with doc.tag('div', klass=klass):
      with doc.tag(f'h{level}', id=id):
        doc.asis(f"{content.sectionNumber}&nbsp{content.sectionTitle}")
      if self._standard_section(content.text):
        #print(f"STD1 {content.text}")
        name = self._standard_section_name(content.text)
        #print(f"STD2 {name}")
        content.text = self._generate_standard_section(name)
        #print(f"STD3 {content.text}")
      doc.asis(self._translate_references(content.text))
      for id in content.contentChildIds:
        content = next((x for x in self.protocol_document_version.contents if x.id == id), None)
        self._content_to_html(content, doc)

  def _translate_references(self, content_text):
    soup = BeautifulSoup(content_text, 'html.parser')
    for ref in soup(['usdm:ref']):
      #print(f"TRA: {ref}")
      attributes = ref.attrs
      #print(f"TRB: {attributes}")
      try:
        if 'namexref' in attributes:
          instance = cross_references.get(attributes['klass'], attributes['namexref'])
        else:
          instance = cross_references.get_by_id(attributes['klass'], attributes['id'])
        try:
          #print(f"TR1: {instance.id}")
          value = getattr(instance, attributes['attribute'])
          #print(f"TR2: {value}")
          translated_text = self._translate_references(value)
          #print(f"TR3: {translated_text}")
          ref.replace_with(translated_text)
        except:
          ref.replace_with("***** Failed to translate reference, attribute not found *****")
      except:
        ref.replace_with("***** Failed to translate reference, instance not found *****")
    return str(soup)
  
  def _standard_section(self, text):
    soup = BeautifulSoup(text, 'html.parser')
    for section in soup(['usdm:section']):
      return True
    return False
  
  def _standard_section_name(self, text):  
    soup = BeautifulSoup(text, 'html.parser')
    for section in soup(['usdm:section']):
      attributes = section.attrs
      if 'name' in attributes:
        return attributes['name'].upper()
      else:
        return None
    return None

  def _generate_standard_section(self, name):
    #print(f"GSS: {name}")   
    if name == "M11-TITLE-PAGE":
      return self._generate_m11_title_page()
    elif name == "M11-INCLUSION":
      return self._generate_m11_criteria("C25532")
    elif name == "M11-EXCLUSION":
      return self._generate_m11_criteria("C25370")
    elif name == "M11-OBJECTIVE-ENDPOINTS":
      return self._generate_m11_objective_endpoints()
    else:
      return f"Unrecognized standard content name {name}"

  def _generate_m11_title_page(self):
    #print(f"M11 TP:")
    doc = Doc()
    with doc.tag('table'):
      self._generate_m11_title_page_entry(doc, 'Protocol Full Title:', f'<usdm:ref klass="StudyProtocolDocumentVersion" id="{self.protocol_document_version.id}" attribute="officialTitle"/>')
      self._generate_m11_title_page_entry(doc, 'Protocol Number:', f'<usdm:ref klass="StudyIdentifier" id="{self._study_identifier().id}" attribute="studyIdentifier"/>')
      self._generate_m11_title_page_entry(doc, 'Version:>', f'<usdm:ref klass="StudyVersion" id="{self.study_version.id}" attribute="studyVersion"/>')
      self._generate_m11_title_page_entry(doc, 'Amendment Number:', f'<usdm:ref klass="StudyAmendment" id="{self._amendment().id}" attribute="number"/>')
      self._generate_m11_title_page_entry(doc, 'Amendment Scope:', '')
      self._generate_m11_title_page_entry(doc, 'Compound Number(s):', '')
      self._generate_m11_title_page_entry(doc, 'Compound Name(s):', '')
      self._generate_m11_title_page_entry(doc, 'Trial Phase:', f'<usdm:ref klass="Code" id="{self.study_version.studyPhase.standardCode.id}" attribute="decode"/>')
      self._generate_m11_title_page_entry(doc, 'Acronym:', f'<usdm:ref klass="StudyVersion" id="{self.study_version.id}" attribute="studyAcronym"/>')
      self._generate_m11_title_page_entry(doc, 'Short Title:', f'<usdm:ref klass="StudyProtocolDocumentVersion" id="{self.protocol_document_version.id}" attribute="briefTitle"/>')
      self._generate_m11_title_page_entry(doc, 'Sponsor Name and Address:', f'<usdm:ref klass="Organization" id="{self._organization().id}" attribute="name"/><br/><usdm:ref klass="Address" id="{self._organization_address().id}" attribute="text"/>')
    result = doc.getvalue()
    #print(f"DOC: {result}")
    return result
  
  def _generate_m11_criteria(self, type):
    #print(f"M11 TP:")
    heading = { 
      'C25532': "Patients may be included in the study only if they meet <strong>all</strong> the following criteria:",
      'C25370': "Patients may be excluded in the study for <strong>any</strong> of the following reasons:",
    }
    doc = Doc()
    with doc.tag('p'):
      doc.asis(heading[type])  
    with doc.tag('table'):
      for criterion in self._criteria(type):
        self._generate_m11_critieria_entry(doc, criterion['identifier'], criterion['text'])
    return doc.getvalue()

  def _generate_m11_objective_endpoints(self):
    #print(f"M11 TP:")
    doc = Doc()
    with doc.tag('table'):
      for item in self._objective_endpoints():
        self._generate_m11_objective_endpoints_entry(doc, item['objective'], item['endpoints'])
    return doc.getvalue()

  def _generate_m11_critieria_entry(self, doc, number, entry):
    with doc.tag('tr'):
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(number)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(entry)

  def _generate_m11_objective_endpoints_entry(self, doc, objective, endpoints):
    with doc.tag('tr'):
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(objective)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        for endpoint in endpoints:
          with doc.tag('p'):
            doc.asis(endpoint)

  def _generate_m11_title_page_entry(self, doc, title, entry):
    with doc.tag('tr'):
      with doc.tag('th', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(title)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(entry)

  def _study_identifier(self):
    identifiers = self.study_version.studyIdentifiers
    for identifier in identifiers:
      if identifier.studyIdentifierScope.type.code == 'C70793':
        return identifier
    return None
  
  def _organization(self):
    identifier = self._study_identifier()
    return identifier.studyIdentifierScope
  
  def _organization_address(self):
    organization = self._organization()
    return organization.organizationLegalAddress
  
  def _amendment(self):
    amendments = self.study_version.amendments
    return amendments[-1]
    
  def _criteria(self, type):
    results = []
    items = [c for c in self.study_design.studyEligibilityCritieria if c.category.code == type ]
    items.sort(key=lambda d: d.identifier)
    for item in items:
      result = {'identifier': item.identifier, 'text': item.text}
      dictionary = cross_references.get_by_id('SyntaxTemplateDictionary', item.dictionaryId)
      if dictionary:
        result['text'] = self._substitute_tags(result['text'], dictionary)
      results.append(result)
    return results

  def _objective_endpoints(self):
    results = []
    for item in self.study_design.studyObjectives:
      result = {'objective': item.text, 'endpoints': []}
      dictionary = cross_references.get_by_id('SyntaxTemplateDictionary', item.dictionaryId)
      if dictionary:
        result['objective'] = self._substitute_tags(result['objective'], dictionary)
      for endpoint in item.objectiveEndpoints:
        dictionary = cross_references.get_by_id('SyntaxTemplateDictionary', endpoint.dictionaryId)
        ep_text = endpoint.text
        if dictionary:
          ep_text = self._substitute_tags(ep_text, dictionary)
        result['endpoints'].append(ep_text)
      results.append(result)
    return results

  def _substitute_tags(self, text, dictionary):
      tags = re.findall(r'\[([^]]*)\]', text)
      for tag in tags:
        if tag in dictionary.parameterMap:
          map = dictionary.parameterMap[tag]
          text = text.replace(f"[{tag}]", f'<usdm:ref klass="{map["klass"]}" id="{map["id"]}" attribute="{map["attribute"]}"/>')
      return text



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.encounter import Encounter
from usdm_model.transition_rule import TransitionRule

class StudyDesignEncounterSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignEncounters')
      self.items = []
      for index, row in self.sheet.iterrows():
        start_rule = None
        end_rule = None
        xref = self.read_cell_by_name(index, 'xref', default='')
        name = self.read_cell_by_name(index, ['encounterName', 'name'])
        description = self.read_description_by_name(index, ['encounterDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default='')
        type = self.read_cdisc_klass_attribute_cell_by_name('Encounter', 'encounterType', index, ['encounterType', 'type'])
        setting = self.read_cdisc_klass_attribute_cell_by_name('Encounter', 'encounterEnvironmentalSetting', index, ['encounterEnvironmentalSetting', 'environmentalSetting'])
        modes = self.read_cdisc_klass_attribute_cell_multiple_by_name('Encounter', 'encounterContactModes', index, ['encounterContactModes', 'contactModes'])
        start_rule_text = self.read_cell_by_name(index, 'transitionStartRule')
        end_rule_text = self.read_cell_by_name(index, 'transitionEndRule')
        if not start_rule_text == "":
          start_rule = TransitionRule(id=id_manager.build_id(TransitionRule), name=f"ENCOUNTER_START_RULE_{index + 1}", text=start_rule_text)
        if not end_rule_text == "":
          end_rule = TransitionRule(id=id_manager.build_id(TransitionRule), name=f"ENCOUNTER_START_RULE_{index + 1}", text=end_rule_text)
        try:
          item = Encounter(
            id=id_manager.build_id(Encounter), 
            name=name,
            description=description,
            label=label,
            type=type, 
            encounterEnvironmentalSetting=setting,
            encounterContactModes=modes,
            transitionStartRule=start_rule,
            transitionEndRule=end_rule
          )
        except Exception as e:
          self._general_error(f"Failed to create Encounter object, exception {e}")
          self._traceback(f"{traceback.format_exc()}")
        else:
          self.items.append(item)
          cross_ref = xref if xref else name
          cross_references.add(cross_ref, item)     
      self.double_link(self.items, 'previousEncounterId', 'nextEncounterId')   
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.study_design_encounter_sheet.study_design_encounter_sheet import StudyDesignEncounterSheet
from usdm_model.code import Code

def test_create(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['RuleId_1', 'EncounterId_1', 'RuleId_2', 'EncounterId_2', 'EncounterId_3']
  codes = []
  for index in range(1,10):
    codes.append(Code(id=f'Code{index}', code='code', codeSystem='codesys', codeSystemVersion='3', decode=f"label{index}"))
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect = codes
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['E1', 'Encounter 1', 'Encounter One', 'visit', 'CLINIC', 'in Person', 'start rule', '' ], 
    ['E2', 'Encounter 2', 'Encounter Two', 'visit', 'CLINIC', 'in Person', '', 'end rule'], 
    ['E3', 'Encounter 3', 'Encounter Three', 'visit', 'HOME', 'TELEPHONE CALL', '', '']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['xref', 'encounterName', 'encounterDescription', 'encounterType', 'encounterEnvironmentalSetting', 'encounterContactModes', 'transitionStartRule',	'transitionEndRule'])
  encounters = StudyDesignEncounterSheet("")
  assert len(encounters.items) == 3
  assert encounters.items[0].id == 'EncounterId_1'
  assert encounters.items[0].name == 'Encounter 1'
  assert encounters.items[1].id == 'EncounterId_2'
  assert encounters.items[1].description == 'Encounter Two'
  assert encounters.items[2].id == 'EncounterId_3'
  assert encounters.items[2].type == codes[6]
  
def test_create_with_label(mocker):
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['RuleId_1', 'EncounterId_1', 'RuleId_2', 'EncounterId_2', 'EncounterId_3']
  codes = []
  for index in range(1,10):
    codes.append(Code(id=f'Code{index}', code='code', codeSystem='codesys', codeSystemVersion='3', decode=f"label{index}"))
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect = codes
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['E1', 'Encounter 1', 'Encounter One', '', 'visit', 'CLINIC', 'in Person', 'start rule', '' ], 
    ['E2', 'Encounter 2', 'Encounter Two', 'Label 2', 'visit', 'CLINIC', 'in Person', '', 'end rule'], 
    ['E3', 'Encounter 3', 'Encounter Three', 'Label 3', 'visit', 'HOME', 'TELEPHONE CALL', '', '']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['xref', 'name', 'description', 'label', 'type', 'encounterEnvironmentalSetting', 'encounterContactModes', 'transitionStartRule',	'transitionEndRule'])
  encounters = StudyDesignEncounterSheet("")
  assert len(encounters.items) == 3
  assert encounters.items[0].id == 'EncounterId_1'
  assert encounters.items[0].name == 'Encounter 1'
  assert encounters.items[0].label == ''
  assert encounters.items[1].id == 'EncounterId_2'
  assert encounters.items[1].description == 'Encounter Two'
  assert encounters.items[2].id == 'EncounterId_3'
  assert encounters.items[2].type == codes[6]
  assert encounters.items[2].label == 'Label 3'

def test_create_empty(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['xref', 'encounterName', 'encounterDescription', 'encounterType', 'encounterEnvironmentalSetting', 'encounterContactModes', 'transitionStartRule',	'transitionEndRule'])
  encounters = StudyDesignEncounterSheet("")
  assert len(encounters.items) == 0

def test_read_cell_by_name_error(mocker):
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['E1', 'Encounter 1', 'Encounter One', 'CLINIC', 'in Person', 'start rule', '' ]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['xref', 'encounterName', 'encounterDescription', 'encounterEnvironmentalSetting', 'encounterContactModes', 'transitionStartRule',	'transitionEndRule'])
  encounters = StudyDesignEncounterSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignEncounters"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception [Failed to detect column(s) 'encounterType, type' in sheet] raised reading sheet."
  



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
from usdm_model.intercurrent_event import IntercurrentEvent
from usdm_model.analysis_population import AnalysisPopulation
from usdm_model.estimand import Estimand
from usdm_model.investigational_intervention import InvestigationalIntervention
from usdm_model.endpoint import Endpoint

class StudyDesignEstimandsSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignEstimands')
      self.estimands = []
      current = None
      current_ice_name = None
      current_ice_description = None
      for index, row in self.sheet.iterrows():
        e_summary = self.read_cell_by_name(index, "summaryMeasure")
        ap_description = self.read_description_by_name(index, 'populationDescription')
        ice_name = self.read_cell_by_name(index, ['intercurrentEventName', 'name'])
        ice_description = self.read_description_by_name(index, ['intercurrentEventDescription', 'description'])
        ice_label = self.read_cell_by_name(index, 'label', default='')
        ice_strategy = self.read_cell_by_name(index, "intercurrentEventStrategy")
        treatment_xref = self.read_cell_by_name(index, "treatmentXref")
        endpoint_xref = self.read_cell_by_name(index, "endpointXref")
        if not e_summary == "":
          try:
            ap = AnalysisPopulation(id=id_manager.build_id(AnalysisPopulation), name=f"AP_{index+1}", text=ap_description) 
          except Exception as e:
            self._general_error(f"Failed to create AnalysisPopulation object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            try:
              treatment = cross_references.get(InvestigationalIntervention, treatment_xref)
              endpoint = cross_references.get(Endpoint, endpoint_xref)
              treatment_id = treatment.id
              endpoint_id = endpoint.id
              current = Estimand(id=id_manager.build_id(Estimand), summaryMeasure=e_summary, analysisPopulation=ap, treatmentId=treatment_id, variableOfInterestId=endpoint_id, intercurrentEvents=[])
            except Exception as e:
              self._general_error(f"Failed to create Estimand object, exception {e}")
              self._traceback(f"{traceback.format_exc()}")
            else:
              self.estimands.append(current)  
        if current is not None:
          try:
            ice_name = current_ice_name if ice_name == "" else ice_name
            ice_description = current_ice_description if ice_description == "" else ice_description
            ice = IntercurrentEvent(id=id_manager.build_id(IntercurrentEvent), name=ice_name, description=ice_description, label=ice_label, intercurrentEventStrategy=ice_strategy)
            current_ice_name = ice_name
            current_ice_description = ice_description
          except Exception as e:
            self._general_error(f"Failed to create IntercurrentEvent object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            current.intercurrentEvents.append(ice)
        else:
          self._general_error("Failed to add IntercurrentEvent, no Estimand set")
          self._traceback(f"{traceback.format_exc()}")

    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




from typing import List, Union
from .api_base_model import ApiBaseModelWithId
from .analysis_population import AnalysisPopulation
from .intercurrent_event import IntercurrentEvent

class Estimand(ApiBaseModelWithId):
  summaryMeasure: str
  analysisPopulation: AnalysisPopulation
  treatmentId: str
  variableOfInterestId: str
  intercurrentEvents: List[IntercurrentEvent] = []



class CrossRef():

  def __init__(self):
    self.references = {}
    self.identifiers = {}

  def clear(self):
    self.references = {}
    self.identifiers = {}

  def add(self, name, object):
    klass = object.__class__
    key, id_key = self._key(klass, name, object.id)
    if not key in self.references:
      self.references[key] = object
      self.identifiers[id_key] = object
    else:
      pass # Need exception here

  def get(self, klass, name):
    key, id_key = self._key(klass, name, "")
    if key in self.references:
      return self.references[key]
    else:
      return None

  def get_by_id(self, klass, id):
    key, id_key = self._key(klass, "", id)
    if id_key in self.identifiers:
      return self.identifiers[id_key]
    else:
      return None

  def _key(self, klass, name, id):
    if isinstance(klass, str):
      key = f"{klass}.{name}" 
      id_key = f"{klass}.{id}"
    else:
      key = f"{klass.__name__}.{name}"
      id_key = f"{klass.__name__}.{id}"
    return key, id_key
  
cross_references = CrossRef()


from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
import traceback
import pandas as pd
from usdm_model.indication import Indication
from usdm_model.investigational_intervention import InvestigationalIntervention

class StudyDesignIISheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignII')
      self.indications = []
      self.interventions = []
      for index, row in self.sheet.iterrows():
        name = self.read_cell_by_name(index, ["xref", "name"])
        type = self.read_cell_by_name(index, "type")
        description = self.read_description_by_name(index, "description")
        label = self.read_cell_by_name(index, 'label', default="")
        codes = self.read_other_code_cell_multiple_by_name(index, "codes")
        if type.upper() in ["IND", "INDICATION"]:
          try:
            item = Indication(id=id_manager.build_id(Indication), name=name, description=description, label=label, codes=codes)
          except Exception as e:
            self._general_error(f"Failed to create Indication object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            self.indications.append(item)
            cross_references.add(name, item)
        else:
          try:
            item = InvestigationalIntervention(id=id_manager.build_id(InvestigationalIntervention), description=description, codes=codes)
          except Exception as e:
            self._general_error(f"Failed to create InvestigationalIntervention object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            self.interventions.append(item)
            cross_references.add(name, item)        
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")




import logging

class Error():

  ERROR = logging.ERROR
  WARNING = logging.WARNING
  DEBUG = logging.DEBUG
  INFO = logging.INFO
  LABEL = { ERROR: "error", WARNING: "warning", DEBUG: "debug", INFO: "info" }

  def __init__(self, sheet: str, row: int, column: int, message: str, level: int=ERROR):
    self.sheet = sheet
    self.row = row
    self.column = column
    self.message = message
    self.level = level

  def to_log(self):
    level = self.__class__.LABEL[self.level].capitalize()  
    if self.sheet == None:
      return f"{level}: {self.message}"
    elif self.row == None:
      return f"{level} in sheet {self.sheet}: {self.message}"
    else:
      return f"{level} in sheet {self.sheet} at [{self.row},{self.column}]: {self.message}"

  def to_dict(self):
    return { 'sheet': self.sheet, 'row': self.row, 'column': self.column, 'message': self.message, 'level': self.__class__.LABEL[self.level].capitalize() }



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.ct_version_manager import ct_version_manager
from usdm_excel.option_manager import * 
import traceback
import pandas as pd

class ConfigurationSheet(BaseSheet):

  PARAMS_NAME_COL = 0
  PARAMS_VALUE_COL = 1

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='configuration', header=None)
      option_manager.set(Options.PREVIOUS_NEXT, PrevNextOption.NONE)
      option_manager.set(Options.ROOT, RootOption.API_COMPLIANT)
      option_manager.set(Options.DESCRIPTION, "")
      self._process_sheet()
    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def om(self):
    return option_manager
  
  def _process_sheet(self):
    for rindex, row in self.sheet.iterrows():
      name = self.read_cell(rindex, self.PARAMS_NAME_COL).strip().upper()
      value = self.read_cell(rindex, self.PARAMS_VALUE_COL)
      if name == 'CT VERSION':
        parts = value.split('=')
        if len(parts) == 2:
          ct_version_manager.add(parts[0].strip(), parts[1].strip())
      elif name == 'SDR PREV NEXT':
        if value.strip().upper() == 'SDR':
          option_manager.set(Options.PREVIOUS_NEXT, PrevNextOption.NULL_STRING)
      elif name == 'SDR ROOT':
        self._general_warning("The SDR_ROOT option is now deprecated and will be ignored.")
      elif name == 'SDR DESCRIPTION':
        option_manager.set(Options.DESCRIPTION, value)



from typing import Union, List
from .api_base_model import ApiBaseModelWithId
from .study_amendment_reason import StudyAmendmentReason
from .subject_enrollment import SubjectEnrollment

class StudyAmendment(ApiBaseModelWithId):
  number: str
  summary: str
  substantialImpact: bool
  primaryReason: StudyAmendmentReason
  secondaryReasons: List[StudyAmendmentReason] = []
  enrollments: List[SubjectEnrollment]
  previousId: Union[str, None] = None



from usdm_excel.cdisc_ct_library import cdisc_ct_library
from usdm_excel.code_base import CodeBase

class OtherCT(CodeBase):

  def code(self, code, system, version, decode):
    return self._build(code=code, system=system, version=version, decode=decode)
 


from typing import Union
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .alias_code import AliasCode

class GeographicScope(ApiBaseModelWithId):
  type: Code
  code: Union[AliasCode, None] = None



class SoAColumnRows:
  
  EPOCH_ROW = 0
  CYCLE_ROW = 1
  CYCLE_START_ROW = 2
  CYCLE_PERIOD_ROW = 3
  CYCLE_END_RULE_ROW = 4
  TIMING_ROW = 5
  VISIT_LABEL_ROW = 6
  VISIT_WINDOW_ROW = 7

  HEADER_ROW = 8
  FIRST_ACTIVITY_ROW = 9

  ACTIVITY_COL = 0
  CHILD_ACTIVITY_COL = 1
  BC_COL = 2
  FIRST_VISIT_COL = 3



from typing import List, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance

class ScheduleTimeline(ApiBaseModelWithIdNameLabelAndDesc):
  mainTimeline: bool
  entryCondition: str
  scheduleTimelineEntryId: str
  scheduleTimelineExits: List[ScheduleTimelineExit] = []
  scheduleTimelineInstances: List[Union[ScheduledActivityInstance, ScheduledDecisionInstance]] = []



import pytest
import pandas as pd

xfail = pytest.mark.xfail

from usdm_excel.study_design_dictionary_sheet.study_design_dictionary_sheet import StudyDesignDictionarySheet
from usdm_model.api_base_model import ApiBaseModelWithId

def test_create(mocker):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[ApiBaseModelWithId(id="1"), ApiBaseModelWithId(id="2"), ApiBaseModelWithId(id="3"), ApiBaseModelWithId(id="4"), ApiBaseModelWithId(id="5")]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['DictionaryId_1', 'DictionaryId_2', 'DictionaryId_3']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Dictionary 1', 'Dictionary One',   'Label One',   'Key 1', 'Klass 1', 'Id 1', 'Attribute 1'], 
    ['',             '',                 '',            'Key 2', 'Klass 2', 'Id 2', 'Attribute 2'], 
    ['Dictionary 2', 'Dictionary Two',   'Label Two',   'Key 3', 'Klass 3', 'Id 3', 'Attribute 3'], 
    ['Dictionary 3', 'Dictionary Three', 'Label Three', 'Key 4', 'Klass 4', 'Id 4', 'Attribute 4'], 
    ['',             '',                 '',            'Key 5', 'Klass 5', 'Id 5', 'Attribute 5']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'xref', 'attribute'])
  dictionaries = StudyDesignDictionarySheet("")
  assert len(dictionaries.items) == 3
  assert dictionaries.items[0].id == 'DictionaryId_1'
  assert dictionaries.items[0].name == 'Dictionary 1'
  assert dictionaries.items[0].description == 'Dictionary One'
  assert dictionaries.items[0].label == 'Label One'
  assert dictionaries.items[0].parameterMap['Key 2']['klass'] == 'Klass 2'
  assert list(dictionaries.items[1].parameterMap.keys()) == ['Key 3']
  assert list(dictionaries.items[2].parameterMap.keys()) == ['Key 4', 'Key 5']
  
def test_create_empty(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'xref', 'attribute'])
  dictionaries = StudyDesignDictionarySheet("")
  assert len(dictionaries.items) == 0

def test_read_cell_by_name_error(mocker):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Dictionary 1', 'Dictionary One',   'Label One',   'Key 1', 'Klass 1', 'Attribute 1']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'key', 'class', 'attribute'])
  dictionaries = StudyDesignDictionarySheet("")
  mock_error.assert_called()
  assert call_parameters == [
    ("dictionaries", 1, -1, "Error reading cell 'xref'", 10),
    ('dictionaries', None, None, "Unable to resolve dictionary reference klass: 'Klass 1', name: '', attribute 'Attribute 1'", 30)
  ]
  



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.cross_ref import cross_references
from usdm_excel.id_manager import id_manager
from usdm_model.objective import Objective
from usdm_model.endpoint import Endpoint
from usdm_model.syntax_template_dictionary import SyntaxTemplateDictionary

import traceback
import re

class StudyDesignObjectiveEndpointSheet(BaseSheet):

  def __init__(self, file_path):
    try:
      super().__init__(file_path=file_path, sheet_name='studyDesignOE')
      self.objectives = []
      current = None
      for index, row in self.sheet.iterrows():
        o_text = self.read_cell_by_name(index, 'objectiveText') 
        ep_name = self.read_cell_by_name(index, ['endpointXref', 'endpointName'])
        ep_description = self.read_description_by_name(index, 'endpointDescription')
        ep_label = self.read_cell_by_name(index, ["endpointLabel"], default='') 
        ep_text = self.read_cell_by_name(index, 'endpointText') 
        ep_purpose = self.read_cell_by_name(index, ['endpointPurposeDescription', 'endpointPurpose'], default='None provided')
        ep_level = self.read_cdisc_klass_attribute_cell_by_name('Endpoint', 'endpointLevel', index, "endpointLevel")
        ep_dictionary_name = self.read_cell_by_name(index, 'endpointDictionary', default='')
        self._validate_references(index, 'endpointText', ep_text, ep_dictionary_name)

        if o_text:
          o_name = self.read_cell_by_name(index, ["objectiveXref", "objectiveName"]) 
          o_description = self.read_description_by_name(index, 'objectiveDescription')
          o_label = self.read_cell_by_name(index, ["objectiveLabel"], default='') 
          o_level = self.read_cdisc_klass_attribute_cell_by_name('Objective', 'objectiveLevel', index, "objectiveLevel")
          o_dictionary_name = self.read_cell_by_name(index, 'objectiveDictionary', default='')
          self._validate_references(index, 'objectiveText', o_text, o_dictionary_name)
          try:
            dictionary_id = self._get_dictionary_id(o_dictionary_name)
            current = Objective(id=id_manager.build_id(Objective),
              instanceType="OBJECTIVE",
              name=o_name,
              description=o_description, 
              label=o_label,
              text=o_text,
              level=o_level,
              objectiveEndpoints=[],
              dictionaryId=dictionary_id
            )
          except Exception as e:
            self._general_error(f"Failed to create Objective object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            self.objectives.append(current)
            cross_references.add(o_name, current)
        if current is not None:
          try:
            dictionary_id = self._get_dictionary_id(ep_dictionary_name)
            ep = Endpoint(id=id_manager.build_id(Endpoint),
              instanceType="ENDPOINT",
              name=ep_name,
              description=ep_description,
              label=ep_label,
              text=ep_text, 
              purpose=ep_purpose, 
              level=ep_level,
              dictionaryId=dictionary_id
            )  
          except Exception as e:
            self._general_error(f"Failed to create Endpoint object, exception {e}")
            self._traceback(f"{traceback.format_exc()}")
          else:
            current.objectiveEndpoints.append(ep)
            cross_references.add(ep_name, ep)
        else:
          self._general_error("Failed to add Endpoint, no Objective set")

    except Exception as e:
      self._general_error(f"Exception [{e}] raised reading sheet.")
      self._traceback(f"{traceback.format_exc()}")

  def _validate_references(self, row, column_name, text, dictionary_name):
    if dictionary_name:
      column = self.column_present(column_name)
      dictionary = cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if not dictionary:
        self._warning(row, column, f"Dictionary '{dictionary_name}' not found")
        return
      tags = re.findall(r'\[([^]]*)\]',text)
      for tag in tags:
        if not tag in dictionary.parameterMap:
          self._warning(row, column, f"Failed to find '{tag}' in dictionary '{dictionary_name}'")

  def _get_dictionary_id(self, dictionary_name):
    if dictionary_name:
      dictionary = cross_references.get(SyntaxTemplateDictionary, dictionary_name)
      if dictionary:
        return dictionary.id
      else:
        self._general_error(f"Unable to find dictionary with name '{dictionary_name}'")
    return None



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.id_manager import id_manager
from usdm_excel.cross_ref import cross_references
from usdm_excel.study_soa_sheet.soa_column_rows import SoAColumnRows
from usdm_model.activity import Activity as USDMActivity
from usdm_model.biomedical_concept_surrogate import BiomedicalConceptSurrogate
from usdm_excel.cdisc_biomedical_concept import cdisc_bc_library
from usdm_model.procedure import Procedure
from usdm_model.schedule_timeline import ScheduleTimeline

class Activity():
  
  def __init__(self, parent, row_index):
    self.parent = parent
    self.row_index = row_index
    self.usdm_biomedical_concept_surrogates = []
    self.usdm_biomedical_concepts = []
    self.name = parent.read_cell(row_index, SoAColumnRows.CHILD_ACTIVITY_COL)
    self._bcs, self._prs, self._tls = self._get_observation_cell(row_index, SoAColumnRows.BC_COL)
    self.parent._debug(row_index, SoAColumnRows.BC_COL, f"Activity {self.name} read. BC: {self._bcs}, PR: {self._prs}, TL: {self._tls}")
    self.usdm_activity = self._as_usdm()
    
  def _as_usdm(self):
    surrogate_bc_items = []
    full_bc_items = []
    procedures = []
    for bc in self._bcs:
      if cdisc_bc_library.exists(bc):
        full_bc = cdisc_bc_library.usdm(bc)
        full_bc_items.append(full_bc.id)
        self.usdm_biomedical_concepts.append(full_bc)
      else:
        surrogate = self._to_bc_surrogates(bc)
        surrogate_bc_items.append(surrogate.id)
        self.usdm_biomedical_concept_surrogates.append(surrogate)
    timelineId = ""
    if len(self._tls) > 0:
      timeline = cross_references.get(ScheduleTimeline, self._tls[0])
      if timeline:
        timelineId = timeline.id
      else:
        timelineId = None
        self.parent._general_error(f"Unable to find timeline with name '{self._tls[0]}'")
    for procedure in self._prs:
      ref = cross_references.get(Procedure, procedure)
      if ref is not None:
        procedures.append(ref)
      else:
        self.parent._warning(self.row_index, SoAColumnRows.BC_COL, f"Cross reference error for procedure {procedure}, not found")
    activity = cross_references.get(Activity, self.name)
    if activity is None:
      activity = USDMActivity(
        id=id_manager.build_id(Activity),
        name=self.name,
        description=self.name,
        definedProcedures=procedures,
        activityIsConditional=False,
        activityIsConditionalReason="",
        biomedicalConceptIds=full_bc_items,
        bcCategoryIds=[],
        bcSurrogateIds=surrogate_bc_items,
        activityTimelineId=timelineId
      )
    else:
      activity.definedProcedures = procedures
      activity.biomedicalConceptIds = full_bc_items
      activity.bcSurrogateIds = surrogate_bc_items
      activity.activityTimelineId = timelineId
    return activity
  
  def _to_bc_surrogates(self, name):
    return BiomedicalConceptSurrogate(
      id=id_manager.build_id(BiomedicalConceptSurrogate),
      name=name,
      description=name,
      label=name,
      bcSurrogateReference='None set'
    )
  
  def _get_observation_cell(self, row_index, col_index):
    bcs = []
    prs = []
    tls = []
    if not self.parent.cell_empty(row_index, col_index):
      value = self.parent.read_cell(row_index, col_index)
      outer_parts = value.split(',')
      for outer_part in outer_parts:
        parts = outer_part.split(':')
        if parts[0].strip().upper() == "BC":
          bcs.append(parts[1].strip())
        elif parts[0].strip().upper() == "PR":
          prs.append(parts[1].strip())
        elif parts[0].strip().upper() == "TL":
          tls.append(parts[1].strip())
        else:
          pass
    return bcs, prs, tls



from typing import List, Union
from .api_base_model import ApiBaseModelWithIdAndName

class NarrativeContent(ApiBaseModelWithIdAndName):
  sectionNumber: str
  sectionTitle: str
  text: Union[str, None] = None
  contentChildIds: List[str] = []



from typing import Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .transition_rule import TransitionRule

class StudyElement(ApiBaseModelWithIdNameLabelAndDesc):
  transitionStartRule: Union[TransitionRule, None] = None
  transitionEndRule: Union[TransitionRule, None] = None



from typing import Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class Procedure(ApiBaseModelWithIdNameLabelAndDesc):
  procedureType: str
  code: Code
  procedureIsConditional: bool
  procedureIsConditionalReason: Union[str, None] = None



import pytest
import pandas as pd

from src.usdm_excel.study_soa_v2_sheet.study_soa_v2_sheet import StudySoAV2Sheet
from usdm_excel.study_soa_sheet.timepoint_type import TimepointType
from usdm_excel.study_soa_sheet.window_type import WindowType
from usdm_excel.iso_8601_duration import ISO8601Duration
from src.usdm_excel.base_sheet import BaseSheet

def test_timepoint_type(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  data = { 'col_1': [ 'A:', 'P: 2D ', 'P2:2D', 'C:', 'N3: 3 weeks', 'N: +2 days', 'P: -3 hrs' ] }
  mock_read.return_value = pd.DataFrame.from_dict(data)
  parent = BaseSheet("", "")
  test_data = [
    (0,0,'anchor', 0, None, ISO8601Duration.ZERO_DURATION),
    (1,0,'previous', -1, "2D", "P2D"),
    (2,0,'previous', -2, "2D", "P2D"),
    (3,0,'cycle start', 1, None, ISO8601Duration.ZERO_DURATION),
    (4,0,'next', 3, "3 weeks", "P3W"),
    (5,0,'next', 1, "+2 days", "P2D"),
    (6,0,'previous', -1, "-3 hrs", "PT3H"),
  ]
  for index, test in enumerate(test_data):
    item = TimepointType(parent, test[0], test[1])
    assert(item.timing_type) == test[2]
    assert(item.relative_ref) == test[3]
    assert(item.description) == test[4]
    assert(item.value) == test[5]

def test_synthetic_timepoint_type():
  item = TimepointType(None, 0, 0)
  assert(item.timing_type) == None
  assert(item.relative_ref) == 0
  assert(item.description) == None
  assert(item.value) == ISO8601Duration.ZERO_DURATION

def test_timepoint_type_error(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  data = { 'col_1': [ 'A', 'P: 2 ', '', 'P2: 2 decades' ] }
  mock_read.return_value = pd.DataFrame.from_dict(data)
  parent = BaseSheet("", "Sheet X")
  test_data = [
    (0,0,"Could not decode the timing value, no ':' detected in 'A'"),
    (1,0,"Could not decode the duration value, no value and units found in '2'"),
    (2,0,"Could not decode the timing value, cell was empty"),
    (3,0,"Could not decode the duration value '2 decades'"),
  ]
  for index, test in enumerate(test_data):
    mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
    item = TimepointType(parent, test[0], test[1])
    mock_error.assert_called()
    assert mock_error.call_args[0][0] == "Sheet X"
    assert mock_error.call_args[0][1] == test[0] + 1
    assert mock_error.call_args[0][2] == test[1] + 1
    assert mock_error.call_args[0][3] == test[2]

def test_window_type(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  data = { 'col_1': [ '1..1 Days', '-1..1 days', '-1 .. 1 weeks' ] }
  mock_read.return_value = pd.DataFrame.from_dict(data)
  parent = BaseSheet("", "")
  test_data = [
    (0,0,'1..1 Days', "P1D", "P1D"),
    (1,0,'-1..1 days', "P1D", "P1D"),
    (2,0,'-1 .. 1 weeks', "P1W", "P1W"),
  ]
  for index, test in enumerate(test_data):
    item = WindowType(parent, test[0], test[1])
    assert(item.description) == test[2]
    assert(item.lower) == test[3]
    assert(item.upper) == test[4]

def test_synthetic_window_type():
  item = WindowType(None, 0, 0)
  assert(item.description) == None
  assert(item.lower) == None
  assert(item.upper) == None

def test_window_type_error(mocker):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  data = { 'col_1': [ '1.. Days', '-1.1 days', '-1 .. 1', '.. 1 Weeks', '' ] }
  mock_read.return_value = pd.DataFrame.from_dict(data)
  parent = BaseSheet("", "Sheet X")
  test_data = [
    (0,0,"Could not decode the window value, not all required parts detected in '1.. Days'"),
    (1,0,"Could not decode the window value, not all required parts detected in '-1.1 days'"),
    (2,0,"Could not decode the window value, not all required parts detected in '-1 .. 1'"),
    (3,0,"Could not decode the window value, not all required parts detected in '.. 1 Weeks'")
  ]
  for index, test in enumerate(test_data):
    mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
    item = WindowType(parent, test[0], test[1])
    mock_error.assert_called()
    assert mock_error.call_args[0][0] == "Sheet X"
    assert mock_error.call_args[0][1] == test[0] + 1
    assert mock_error.call_args[0][2] == test[1] + 1
    assert mock_error.call_args[0][3] == test[2]



import pytest
import pandas as pd

from src.usdm_excel.base_sheet import BaseSheet

xfail = pytest.mark.xfail

from usdm_excel.study_design_content_sheet.study_design_content_sheet import StudyDesignContentSheet

def test_create(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['1',     '',         'Section 1',     'Text 1'], 
    ['1.1',   'SET NAME', 'Section 1.1',   'Text 1.1'], 
    ['1.2',   '',         'Section 1.2',   'Text 1.2'], 
    ['1.2.1', '',         'Section 1.2.1', 'Text 1.2.1'], 
    ['2',     '',         'Section 2',     'Text 2'], 
    ['2.1',   '',         'Section 2.1',   'Text 2.1'], 
    ['3',     '',         'Section 3',     'Text 3'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  content = StudyDesignContentSheet("")
  assert len(content.items) == 8
  assert content.items[0].name == 'ROOT'
  assert content.items[1].id == 'Content_2'
  assert content.items[1].name == 'SECTION 1'
  assert content.items[1].sectionNumber == '1'
  assert content.items[1].sectionTitle == 'Section 1'
  assert content.items[1].text == '<div>Text 1</div>'
  assert content.items[1].contentChildIds == ['Content_3', 'Content_4']
  assert content.items[2].name == 'SET NAME'
  assert content.items[3].name == 'SECTION 1.2'
  assert content.items[4].name == 'SECTION 1.2.1'
  assert content.items[5].name == 'SECTION 2'
  assert content.items[6].name == 'SECTION 2.1'
  assert content.items[7].id == 'Content_8'
  assert content.items[7].name == 'SECTION 3'
  
def test_create_4_levels(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['1',       '',         'Section 1',       'Text 1'], 
    ['1.1',     'SET NAME', 'Section 1.1',     'Text 1.1'], 
    ['1.2',     '',         'Section 1.2',     'Text 1.2'], 
    ['1.2.1',   '',         'Section 1.2.1',   'Text 1.2.1'], 
    ['1.2.1.1', '',         'Section 1.2.1.1', 'Text 1.2.1.1'], 
    ['2',       '',         'Section 2',       'Text 2'], 
    ['3',       '',         'Section 3',       'Text 3'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  content = StudyDesignContentSheet("")
  assert len(content.items) == 8
  assert content.items[0].name == 'ROOT'
  assert content.items[4].id == 'Content_5'
  assert content.items[4].name == 'SECTION 1.2.1'
  assert content.items[4].sectionNumber == '1.2.1'
  assert content.items[4].sectionTitle == 'Section 1.2.1'
  assert content.items[4].text == '<div>Text 1.2.1</div>'
  assert content.items[4].contentChildIds == ['Content_6']
  assert content.items[5].id == 'Content_6'
  assert content.items[5].name == 'SECTION 1.2.1.1'
  assert content.items[5].sectionNumber == '1.2.1.1'
  assert content.items[5].sectionTitle == 'Section 1.2.1.1'
  assert content.items[5].text == '<div>Text 1.2.1.1</div>'
  assert content.items[5].contentChildIds == []

def test_create_standard_section(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['1',       '',         'Section 1',       'Text 1'], 
    ['1.1',     'SET NAME', 'Section 1.1',     'Text 1.1'], 
    ['1.2',     '',         'Section 1.2',     '<usdm:section name="m11-title">'], 
    ['1.2.1',   '',         'Section 1.2.1',   '<div><usdm:section name="m11-title"></div>'], 
    ['2',       '',         'Section 2',       'Text 2'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  content = StudyDesignContentSheet("")
  assert len(content.items) == 6
  assert content.items[0].text == ''
  assert content.items[1].text == '<div>Text 1</div>'
  assert content.items[2].text == '<div>Text 1.1</div>'
  assert content.items[3].text == '<div><usdm:section name="m11-title"></div>'
  assert content.items[4].text == '<div><usdm:section name="m11-title"></div>'
  assert content.items[5].text == '<div>Text 2</div>'

def test_create_invalid_levels(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['1',         '',         'Section 1',       'Text 1'], 
    ['1.1',       'SET NAME', 'Section 1.1',     'Text 1.1'], 
    ['1.2',       '',         'Section 1.2',     'Text 1.2'], 
    ['1.2.1',     '',         'Section 1.2.1',   'Text 1.2.1'], 
    ['1.2.1.1.4', '',         'Section 1.2.1.1', 'Text 1.2.1.1'], 
    ['2',         '',         'Section 2',       'Text 2'], 
    ['3',         '',         'Section 3',       'Text 3'], 
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  StudyDesignContentSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignContent"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception [] raised reading sheet."

def test_create_empty(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
  content = StudyDesignContentSheet("")
  assert len(content.items) == 1

def test_read_cell_by_name_error(mocker):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['1', 'Section 1', 'Text 1']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'sectionTitle', 'text'])
  content = StudyDesignContentSheet("")
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignContent"
  assert mock_error.call_args[0][1] == 1
  assert mock_error.call_args[0][2] == -1
  assert mock_error.call_args[0][3] == "Error reading cell 'name'"
  
